#include "MobileDevice.h"
#include "Shell.h"
#include "UUTextParser.h"
#include <vector>
#include "md5.h"
#include "UUErrorCodes.h"

// TagLib
#include "tag.h"
#include "tagunion.h"
#include "fileref.h"

using namespace std;

//---

#define kMaxDevices			8
#define kDirDummySize		(100*kUU_1KB)

#define kDT_default			0
#define kDT_copyFlat		(1L << 0)
#define kDT_copyMerge		(1L << 1)
#define kDT_convertPlists	(1L << 2)
#define kDT_convertPng		(1L << 3)

//---
// file system item returned as binary output

struct bSEItem
{
	int						flags;
	int						isDir;
	unsigned int			size;
	string512				path;
	string256				manifestInfo;
	string64				md5Date;
	
	// 550002 and up
	double					soundDuration;
	string256				soundTitle;
	string256				soundArtist;
};

typedef struct bSEItem bSEItem;

// data transfer item

struct bSEItemDT
{
	struct shell_state		*srcSH;
	string512				srcPath;
	bool					srcIsDir;
	unsigned int			srcSize;
	
	struct shell_state		*dstSH;
	string512				dstPath;
};

typedef struct bSEItemDT bSEItemDT;

// header returned as binary output

struct bSEHdr
{
	int						result;
	unsigned int			numItems;
};

typedef struct bSEHdr bSEHdr;

//---
// device structure

struct bSEDevice
{
	struct shell_state		*sh;
	string256				serialNo;
	unsigned int			connID;
	
	bool					didPair;
	bool					didStartSession;
	bool					didStartAfc;
	bool					didStartAfc2;
	bool					didOpenConn;
};

typedef struct bSEDevice bSEDevice;

// file structure; device or disk

struct bSEFile
{
	string512				path;
	struct shell_state		*sh;
	
	FILE					*fp;
	afc_file_ref			ref;
};

typedef struct bSEFile bSEFile;

// directory structure; device or disk

struct bSEDir
{
	string512				path;
	struct shell_state		*sh;
	
	struct afc_directory	*dir;
	
	std::vector<bSEItem>	*PIL;
};

typedef struct bSEDir bSEDir;

//---

bSEDevice				gDEV[kMaxDevices];

int						gInstanceID;
bool					gMultiDevice;
char					gSerialNo[256];
bool					gOneShot;
char					gMsgStr[8192];
bool					gVerbose;
bool					gShowProgress;
bool					gCheckAbort;
double					gTimeOut;
bool					gRecoveryModesOK;
bool					gSkipStdOut;
bool					gUseManifest;
bool					gAppChange;
bool					gGetMD5Date;
bool					gGetSoundInfo;
string512				gManifestPath;
string512				gPath1;
string512				gSkipListPath;

string512				gConfigPath;
string512				gInputPath;
string512				gOutputPath;
string512				gBinOutputPath;
string512				gProgressPath;
string512				gProgressMsgPath;
string512				gAbortPath;
string2K				gDummyMP3;
int						gDummyMP3Size;

UUTextParser			gConfigTP;
UUTextParser			gOutTP;
UUTextParser			gSkipTP;

int						gWaitForRecoveryIndex;

unsigned long long		gTotalDataSize;
unsigned long long		gProcessedDataSize;
unsigned int			gTotalCount;
unsigned int			gProcessedCount;

bool					gDataTransferPreflight;
std::vector<bSEItemDT>	gDataTransferIL;

//---

void bSE_fileSet(struct shell_state *sh, char *path, bSEFile *f);
int bSE_fileOpen(bSEFile *f, bool readOnly);
void bSE_fileClose(bSEFile *f);
int bSE_fileRead(bSEFile *f, unsigned char *buf, unsigned int size);
int bSE_fileWrite(bSEFile *f, unsigned char *buf, unsigned int size);
int bSE_fileSeek(bSEFile *f, unsigned long long offset1, unsigned long long offset2);
int bSE_fileDelete(bSEFile *f, bool isDir);
unsigned long bSE_fileSize(bSEFile *f, int *pfailed = nil);
void bSE_getMD5Date(bSEFile *f, unsigned long size, char *md5Date);
void bSE_getSoundInfo(bSEFile *f, bSEItem *pI);
void bSE_getSoundInfo_AAC(bSEFile *f, bSEItem *pI);
void bSE_getSoundInfo_MP3(bSEFile *f, bSEItem *pI);

void bSE_dirSet(struct shell_state *sh, char *path, bSEDir *d);
int bSE_dirCreate(bSEDir *d);
int bSE_dirOpen(bSEDir *d);
void bSE_dirClose(bSEDir *d);
int bSE_dirNextPath(bSEDir *d, char *nextAbsPath, char *nextName);
int bSE_dirMakeEmpty(bSEDir *d);
int bSE_dirDelete(bSEDir *d);

void bSE_getSerialNo(struct am_device *dev, char *serialNo);
void bSE_getDeviceInfo(struct shell_state *sh);
int bSE_readFile(struct shell_state *sh, char *path, UUTextParser *fileTP);
int bSE_writeFile(struct shell_state *sh, char *path, char *data, int dataSize);
int bSE_writeBinaryOutput(bSEHdr *hdr, std::vector<bSEItem> *PIL);

int bSE_dataTransfer(struct shell_state *srcSH, char *srcPath, struct shell_state *dstSH, char *dstPath, int transferFlags);

int bSE_dataDelete(struct shell_state *srcSH, char *srcPath);
int bSE_dataRename(struct shell_state *sh, char *path, char *newName);
unsigned long long bSE_totalDataSize(struct shell_state *sh, char *path, int transferFlags);
unsigned int bSE_totalCount(struct shell_state *sh, char *path);
int bSE_dataInfo(struct shell_state *sh, char *path, std::vector<bSEItem> *PIL);
int bSE_listDirectory(struct shell_state *srcSH, char *srcPath);
int bSE_listDirectoryDeep(struct shell_state *srcSH, char *srcPath);
int bSE_listDirectoryDeep2(struct shell_state *srcSH, char *srcPath, std::vector<bSEItem> *PIL);

int bSE_copyFile(struct shell_state *srcSH, char *srcPath, struct shell_state *dstSH, char *dstPath, int transferFlags);
int bSE_copyFolder(struct shell_state *srcSH, char *srcPath, struct shell_state *dstSH, char *dstPath, int transferFlags);
int bSE_diskDirExists(char *path);

int bSE_openConnection(bSEDevice *dev, char *forSerialNo, bool useAfc2);

bool bSE_abort(void);
bool bSE_appChange(char *path);

void bSE_updateProgress(int numDone, int numTotal);
void bSE_updateProgressMsg(char *msg);
void bSE_updateDataTransferProgress(unsigned int transferredBytes);
void bSE_updateDataDeleteProgress(void);

int bSE_processShellCommands(struct shell_state *sh, char *cmdFile);
void bSE_deviceCommand(char *command, char *serialNo);

void bSE_updateManifest(struct shell_state *sh);
void bSE_loadManifest(struct shell_state *sh);
bool bSE_getManifestInfo(struct shell_state *sh, char *path, char *manifestInfo);
void bSE_updateManifestTP(UUTextParser *TP, const char *phonePath, const char *text);

errcode mConvertPlist(char *data, int dataSize, UUTextParser *outTP);
errcode mConvertPng(char *pngPath);
errcode mExitRecoveryWithIPhonecomm(char *commPath);
errcode mShellCommand(char *command, char *result, long maxResultSize);
int mGetUnicodeString(char *cs, UniChar *ucs);
errcode mSplitPath(const char *fullPath, char *path, char *name);
errcode mGetSoundInfo_MP3(char *path, bSEItem *pI);
errcode mGetSoundInfo_AAC(char *path, bSEItem *pI);

void dispose_shell(struct shell_state *sh);

//---

COMMAND normal_shell[] = 
{
	{ "help",			sh_help,         			">> help <command> - Display help information on <command>.  No args lists commands." },
	{ "cd",				n_cd,            			">> cd <path> - Change directory to <path>" },
	{ "lcd",			n_lcd,						">> lcd <path> - Change local directory to <path>" },
	{ "ls",				n_ls,            			">> ls [path] - List directory at current working path or [path]" },
	{ "lsx",			n_lsx,           			">> lsx [path] - List directory at current working path or [path] with file sizes" },
	{ "lsd",			n_lsd,           			">> lsd [path] - List directory at current working path or [path]; recursively" },
	{ "mkdir",			n_mkdir,         			">> mkdir <path> - Create directory at <path>" },		
	{ "rmdir",			n_rmdir,         			">> rmdir <path> - Remove directory at <path>" },
	{ "activate",		n_activate,      			">> activate <path> - Activate iPhone with .plist file at <path>" },
	{ "deactivate",		n_deactivate,    			">> deactivate - Deactivate iPhone" },
	{ "readvalue",		n_readvalue,     			">> readvalue <value> - Read <value>.  No args lists knownn values." },
	{ "enterrecovery",	n_enterrecovery, 			">> enterrecovery - Enter recovery mode. **WARNING: You'll need to restore the iPhone." },
	{ "reconnect",		n_reconnect,     			">> reconnect - Exit current shell and reconnect to device." },
	{ "startservice",	n_startservice,  			">> startservice <service> - Starts service <service> on the iPhone.  No args lists services." },
	{ "deviceinfo",		n_deviceinfo,    			">> deviceinfo - Display device info." },
	{ "getfilesize",	n_getfilesize,   			">> getfilesize <path> - Display size of file at <path>" },
	{ "getfile",		n_getfile,       			">> getfile <path> [localpath] - Get file on iPhone at <path> and write it to [localpath]" },
	{ "putfile",		n_putfile,       			">> putfile <localpath> [path] - Put file at <localpath> on iPhone at [path]" },
	{ "fileinfo",		n_fileinfo,      			">> fileinfo <path> - Display info for file at <path>" },
	{ "exit",			n_exit,          			">> exit - Escape to shell.  The other shell, the one whos child i am." },
	{ "lpwd",			n_lpwd,						">> lpwd - Display the current local working directory." },
	{ "pwd",			n_pwd,           			">> pwd - Display the current remote working directory." },
	{ "setafc",			n_setafc,					">> setafc <name string> - Set the name of the afc service to use. ('x' for com.apple.afc2)"},
	{ "copyfile",		n_copyfile,      			">> copyfile <path1> <path2> - copy file on iPhone from <path1> and write it to <path2>" },
	{ "backupfile",		n_backupfile,    			">> backupfile <path1> - back up <path1>xxx file on iPhone to <path1>_org_xxx" },
	{ "unbackupfile",	n_unbackupfile,  			">> unbackupfile <path1> - restore <path1>_org_xxx file on iPhone to <path1>xxx and delete <path1>_org_xxx" },
	{ "restorefile",	n_restorefile,				">> restorefile <path1> - restore <path1>_org_xxx file on iPhone to <path1>xxx" },
	{ "status",			n_status,					">> status - returns some status information from/about the phone" },

	{ "deletePlaceholders",	n_deletePlaceholders,	">>" },
	{ "deleteAll_bSE",	n_deleteAll_bSE,			">>" },
	{ "activateSonOf",	n_activateSonOf,			">>" },

	{ (char *)NULL, (shell_funct *)NULL, (char *)NULL }
};

COMMAND restore_shell[] =
{
	{ "mount",			restore_mount,				">> mount <device> <path> - Mount device at path."},
	{ "partition",		restore_partition,			">> partition <device> - Partition device."},
	{ "erase",			restore_erase,				">> erase <device> - Erase device."},
	{ "ditto",			restore_ditto,				">> ditto <path1> <path2> - Copy file at path1 to path2."},
	{ "umount",			restore_umount,				">> umount <path> - Unmount device from path."},
	{ "FileSystemCheck",restore_filesystemcheck,	">> filesystemcheck <device> - Check filesytem on device"	},
	{ "mkdir",			restore_mkdir,				">> mkdir <path> - Make directory 'path'."},
	{ "force",			restore_force,				">> force <command> - send cmd to phone."},
	{ "help",			sh_help,					">> help <command> - Display help information on <command>.  No args lists commands." },
	{ "exit",			restore_exit,				">> exit - Disconnect and wait for device normal reconnect." },
	{ (char *)NULL, (shell_funct *)NULL, (char *)NULL }
};

COMMAND recovery_shell[] =
{
	{ "restore",		recovery_restore,			">> restore - enter restore mode. "},
	{ "grestore",		recovery_grestore,			">> grestore <restorebundle> or  grestore <...dmg path> <kernelcache... path> - enter restore mode interactively."},
	{ "filecopytophone",recovery_filecopytophone,	">> filecopytophone"},
	{ "serial",			recovery_serial,			">> serial"},
	{ "cmd",			recovery_cmd,				">> cmd <command> - send command to phone."},
	{ "exit",         	recovery_exit,				">> exit - Escape to shell.  The other shell, the one whos child i am." },
	{ "disconnect",     recovery_disconnect,        ">> disconnect - disconnect from shell and await reconnect." },		
	{ "help",			sh_help,					">> help <command> - Display help information on <command>.  No args lists commands." },
	{ (char *)NULL, (shell_funct *)NULL, (char *)NULL }

};

COMMAND dfu_shell[] =
{
	{ "restore",		dfu_restore,				">> restore - enter restore mode. "},
	{ "exit",         	dfu_exit,					">> exit - Escape to shell.  The other shell, the one whos child i am." },
	{ "disconnect",     dfu_disconnect,				">> disconnect - disconnect from shell and await reconnect." },		
	{ "help",			sh_help,					">> help <command> - Display help information on <command>.  No args lists commands." },
	{ (char *)NULL, (shell_funct *)NULL, (char *)NULL }
};

//------------------------------------------------------------------------------------------
#if 0 
#pragma mark -
#endif
//------------------------------------------------------------------------------------------
// list all connected devices and set up a shell object for each one

void listDevices_recovery_connect_callback(am_recovery_device *rdev)
{
	struct shell_state	*sh;

	for (int i = 0; i < kMaxDevices; ++i)
	{
		if (! gDEV[i].sh)
		{
			sh = new shell_state();
			
			if (sh)
			{
				sh->devNo			= i+1;
				sh->dev				= NULL;
				sh->restore_dev		= NULL;
				sh->recovery_dev	= rdev;
				sh->shell_mode		= SHELL_RECOVERY;
				sh->command_array	= recovery_shell;
				sh->remote_path		= "#";
				sh->local_path		= "#";
				sh->prompt_string	= "(iPHUC Recovery) ";
				
				sh->MF				= nil;
				sh->manifestTP		= nil;
				
				gDEV[i].sh = sh;
			}
			
			break;
		}
	}
}

//------------------------------------------------------------------------------------------

void listDevices_dfu_connect_callback(am_recovery_device *rdev)
{
	struct shell_state	*sh;
	
	for (int i = 0; i < kMaxDevices; ++i)
	{
		if (! gDEV[i].sh)
		{
			sh = new shell_state();
			
			if (sh)
			{
				sh->devNo			= i+1;
				sh->dev				= NULL;
				sh->restore_dev		= NULL;
				sh->recovery_dev	= rdev;
				sh->shell_mode		= SHELL_DFU;
				sh->command_array	= dfu_shell;
				sh->remote_path		= "#";
				sh->local_path		= "#";
				sh->prompt_string	= "(iPHUC DFU) ";
			
				sh->MF				= nil;
				sh->manifestTP		= nil;
				
				gDEV[i].sh = sh;
			}
			
			break;
		}
	}
}

//------------------------------------------------------------------------------------------

void listDevices_notification(struct am_device_notification_callback_info *info)
{
	struct am_device	*dev = info->dev;
	unsigned int		msg = info->msg;
	struct shell_state	*sh;
	
	if (msg == ADNCI_MSG_CONNECTED)
	{
		if (AMDeviceConnect(dev))
			goto finish;
			
		for (int i = 0; i < kMaxDevices; ++i)
		{
			if (! gDEV[i].sh)
			{
				sh = new shell_state();
				
				if (sh)
				{
					sh->devNo			= i+1;
					sh->dev				= dev;
					sh->shell_mode		= SHELL_NORMAL;
					sh->command_array	= normal_shell;
					sh->remote_path		= "/";
					sh->local_path		= "/";
					sh->prompt_string	= "(iPHUCMD) ";
				
					sh->MF				= nil;
					sh->manifestTP		= nil;

					gDEV[i].sh = sh;
				}
				
				break;
			}
		}
	}

finish:
	return;
}

//------------------------------------------------------------------------------------------
#if 0 
#pragma mark -
#endif
//------------------------------------------------------------------------------------------

void dfu_connect_callback(am_recovery_device *rdev)
{
	struct shell_state	*sh;
	int					ret;
	
	my_cout("DFU callback: //DFU_MODE//", kForceOut);

	if (! gRecoveryModesOK)
		my_exit();
	
	sh = new shell_state();
	
	sh->devNo			= 0;
	sh->dev				= NULL;
	sh->restore_dev		= NULL;
	sh->recovery_dev	= rdev;
	sh->shell_mode		= SHELL_DFU;
	sh->command_array	= dfu_shell;
	sh->remote_path		= "#";
	sh->local_path		= "#";
	sh->prompt_string	= "(iPHUC DFU) ";
	
	sh->MF				= nil;
	sh->manifestTP		= nil;

	//ret = shell(sh);
	ret = bSE_processShellCommands(sh, gInputPath);
	dispose_shell(sh);
	
	my_cout("DFU callback: shell returned: ", ret);
	my_exit();
}

//------------------------------------------------------------------------------------------

void dfu_disconnect_callback(am_recovery_device *rdev)
{
	my_cout(">> DFU Mode Disconnect.", kForceOut);
}

//------------------------------------------------------------------------------------------

void recovery_connect_callback(am_recovery_device *rdev)
{
	struct shell_state	*sh;
	int					ret;

#if 0 // skip the log...
	my_cout("recovery callback: Logging in restore.log: ", AMRestoreEnableFileLogging("restore.log"));
#endif
	
	my_cout("recovery callback: //RECOVERY_MODE//", kForceOut);
	
	if (! gRecoveryModesOK)
		my_exit();

	sh = new shell_state();
	
	sh->devNo			= 0;
	sh->dev				= NULL;
	sh->restore_dev		= NULL;
	sh->recovery_dev	= rdev;
	sh->shell_mode		= SHELL_RECOVERY;
	sh->command_array	= recovery_shell;
	sh->remote_path		= "#";
	sh->local_path		= "#";
	sh->prompt_string	= "(iPHUC Recovery) ";

	sh->MF				= nil;
	sh->manifestTP		= nil;

	//ret = shell(sh);
	ret = bSE_processShellCommands(sh, gInputPath);
	dispose_shell(sh);
	
	my_cout("recovery callback: shell returned: ", ret);
	my_exit();
}

//------------------------------------------------------------------------------------------

void recovery_disconnect_callback(am_recovery_device *rdev)
{
	my_cout(">> Recovery Mode Disconnect.", kForceOut);
}

//------------------------------------------------------------------------------------------

void notification(struct am_device_notification_callback_info *info)
{
	struct am_device	*dev = info->dev;
	unsigned int		msg = info->msg, connID;
	int					ret;
	struct shell_state	*sh = nil;
	char				serialNo[256], *snStr;
	char				cmdFile[256], outStr[512];
	bool				doDisconnect = false;
	
	if (msg == ADNCI_MSG_CONNECTED)
	{
		my_cout("notification: device attached.");
		ret = AMDeviceConnect(dev);	
		my_cout("AMDeviceConnect: ", ret);

		if (ret)
		{
			my_cout("notification: Could not connect.", kForceOut);
		
			if (gRecoveryModesOK) // check for restore mode
			{
				my_cout(">> Attempting to connect to device in Restore mode.", kForceOut);
				my_cout("notification: //RESTORE_MODE//", kForceOut);
			
				sh = new shell_state();
				
				sh->devNo				= 0;
				sh->dev					= dev;
				sh->restore_dev			= AMRestoreModeDeviceCreate( 0 , AMDeviceGetConnectionID(dev),0);
				sh->restore_dev->port	= socketForPort(sh->restore_dev, 0xf27e);
				
				my_cout(">> Restore Mode Port: ", sh->restore_dev->port);
				
				sh->shell_mode			= SHELL_RESTORE;
				sh->command_array		= restore_shell;
				sh->remote_path			= "#";
				sh->local_path			= "#";
				sh->prompt_string		= "(iPHUC Restore) ";
			
				sh->MF					= nil;
				sh->manifestTP			= nil;

				//ret = shell(sh);
				ret = bSE_processShellCommands(sh, gInputPath);
				dispose_shell(sh);
		
				my_cout("notification: shell returned: ", ret);
				my_exit();
			}
		} 
		else 
		{
			sh = nil;
			
			for (int i = 0; i < kMaxDevices; ++i)
			{
				if (! gDEV[i].sh)
				{
					sh = new shell_state();
					if (sh)
					{
						sh->devNo			= i+1;
						sh->dev				= dev;
						gDEV[i].sh			= sh;
						snStr				= gDEV[i].serialNo;
						gDEV[i].connID		= AMDeviceGetConnectionID(dev);
						
						// build the rest of shell_state
						
						sh->shell_mode		= SHELL_NORMAL;
						sh->command_array	= normal_shell;
						sh->remote_path		= "/";
						sh->local_path		= "/";
						sh->prompt_string	= "(iPHUCMD) ";

						sh->MF				= nil;
						sh->manifestTP		= nil;
					}
					break;
				}
			}
			
			if (sh)
			{
				AMDeviceIsPaired(sh->dev);
				ret = AMDeviceValidatePairing(sh->dev);
				if (ret) 
				{
					sprintf(outStr, "//COMM_ERROR//STATUS_PAIRING_FAILED//%d//ERROR: %ld//", sh->devNo, (long) ret);
					my_cout(outStr, kForceOut);
				}
				else
				{
					ret = AMDeviceStartSession(sh->dev);
					if (ret) 
					{
						sprintf(outStr, "//COMM_ERROR//STATUS_SESSION_START_FAILED//%d//ERROR: %ld//", sh->devNo, (long) ret);
						my_cout(outStr, kForceOut);
					}
					else
					{
						bSE_getSerialNo(dev, serialNo);
						
						if (! gMultiDevice)
							if (gSerialNo[0])
								if (strcmp(gSerialNo, serialNo))
								{
									doDisconnect = true;
									goto finish;
								}
							
						strcpy(snStr, serialNo);
		
						ret = AMDeviceStartService(sh->dev, AMSVC_AFC, &(sh->afch), NULL);	
						if (ret) 
						{
							sprintf(outStr, "//COMM_ERROR//STATUS_AFC_START_FAILED//%d//ERROR: %ld//", sh->devNo, (long) ret);
							my_cout(outStr, kForceOut);
						}
						else
						{
							ret = AFCConnectionOpen(sh->afch, 0, &(sh->conn));
							if (ret) 
							{
								sprintf(outStr, "//COMM_ERROR//STATUS_AFC_OPEN_FAILED//%d//ERROR: %ld//", sh->devNo, (long) ret);
								my_cout(outStr, kForceOut);
							}
							else
							{
								if (gMultiDevice)
								{
									bSE_getDeviceInfo(sh);
									sprintf(cmdFile, "CMD_%s.txt", serialNo);
								}
								else
								{
									sprintf(outStr, "//DEVICE%d: %s//", sh->devNo, serialNo);
									my_cout(outStr, kForceOut);
									strcpy(cmdFile, gInputPath);
								}
								
								my_cout("notification: //NORMAL_MODE//", kForceOut);
								
								if (! bSE_processShellCommands(sh, cmdFile))
								{
									if (sh->shell_mode == SHELL_NORMAL)
										AFCConnectionClose(sh->conn);
									
									my_cout("//OK//", kForceOut);
									my_exit();
								}
							}
						}
					}
				}
			}
		}
	} 
	else if ( msg == ADNCI_MSG_DISCONNECTED ) 
	{
		my_cout("notification: //STATUS_DISCONNECTED//", kForceOut);
		doDisconnect = true;
	}
	else
	{
		my_cout("Waiting...", kForceOut);
	}

finish:
	
	if (doDisconnect)
	{
		connID = AMDeviceGetConnectionID(dev);
		
		for (int i = 0; i < kMaxDevices; ++i)
		{
			if (connID == gDEV[i].connID)
			{
				sprintf(outStr, "device%d; %s", i+1, gDEV[i].serialNo);
				my_cout(outStr, kForceOut);
				
				dispose_shell(gDEV[i].sh); 
				gDEV[i].sh = nil;
				
				MEMSET(&gDEV[i], 0, sizeof(bSEDevice));
			}
		}
	}
	
	return;
}

//------------------------------------------------------------------------------------------

void dispose_shell(struct shell_state *sh)
{
	if (sh)
	{
		if (sh->MF) delete sh->MF;
		sh->MF = nil;
		
		if (sh->manifestTP) delete sh->manifestTP;
		sh->manifestTP = nil;
		
		delete sh;
	}
}

//------------------------------------------------------------------------------------------

void my_cout(const char *str, int val)
{
	string1K	cs;
	
	if (val == kNoVal) // ignore
	{
		if (gVerbose)	
		{
			if (! gSkipStdOut)
				cout << str << endl;

			sprintf(cs, "%s\n", str);
			gOutTP.mAddString(str);
		}
	}
	else if (val == kForceOut)
	{
		if (! gSkipStdOut)
			cout << str << endl;
		
		sprintf(cs, "%s\n", str);
		gOutTP.mAddString(cs);
	}
	else if (val == kForceOutErr)
	{
		if (! gSkipStdOut)
			cout << endl << "//ERROR// " << str << endl;
		
		sprintf(cs, "\n//ERROR// %s\n", str);
		gOutTP.mAddString(cs);
	}
	else if (val)
	{
		if (! gSkipStdOut)
			cout << endl << "//ERROR// " << str << " " << val << endl;
		
		sprintf(cs, "\n//ERROR// %s %d\n", str, val);
		gOutTP.mAddString(cs);
	}
	else if (gVerbose)
	{
		if (! gSkipStdOut)
			cout << str << " " << val << endl;
		
		sprintf(cs, "%s %d\n", str, val);
		gOutTP.mAddString(cs);
	}
	// else blessed silence
}

//------------------------------------------------------------------------------------------

void my_exit(void)
{
	gOutTP.mWriteTextToFile(gOutputPath);
	cout << endl;
	exit(0);
}

//------------------------------------------------------------------------------------------

int main(int argc, char **argv)
{
	struct am_device_notification	*notif; 
	unsigned int					ret;
	SInt32							res;
	string1K						deviceCommand, outStr;
	
	gWaitForRecoveryIndex = -1;

	deviceCommand[0] = 0;
	gOutTP.mSetText(nil, 512L * kUU_1KB);
	
	if (argc > 1)
		gInstanceID = strtol(argv[1], nil, 10);
	if (argc > 2)
		strcpy(deviceCommand, argv[2]);

	sprintf(outStr, 
	"iphucMD by bitSplit(tm) Enterprises\n"
	"based upon the iPhone Dev Team's iphuc\n"
	"#%d %s"
	, gInstanceID
	, deviceCommand
	);
	my_cout(outStr, kForceOut);

	sprintf(gConfigPath,		"iphucMD%dConfig.txt", gInstanceID);
	sprintf(gInputPath,			"iphucMD%dInput.txt", gInstanceID);
	sprintf(gOutputPath,		"iphucMD%dOutput.txt", gInstanceID);
	sprintf(gProgressPath,		"iphucMD%dProgress.txt", gInstanceID);
	sprintf(gProgressMsgPath,	"iphucMD%dProgressMsg.txt", gInstanceID);
	sprintf(gAbortPath,			"iphucMD%dAbort.txt", gInstanceID);
	sprintf(gBinOutputPath,		"iphucMD%dBinOutput.dat", gInstanceID);
	
	gConfigTP.mSetText(gConfigPath);
	
	unlink(gOutputPath);
//	unlink(gProgressPath); caller should write 0 to the file before calling
//	unlink(gProgressMsgPath); caller should write initial message to the file before calling
	unlink(gAbortPath);
	unlink(gBinOutputPath);
	
	if (! deviceCommand[0])
		strcpy(deviceCommand, gConfigTP.mGetString("COMMAND", ""));
		
	strcpy(gSerialNo, gConfigTP.mGetString("SERIAL_NO", ""));
	strcpy(gPath1, gConfigTP.mGetString("PATH1", ""));
	strcpy(gSkipListPath, gConfigTP.mGetString("SKIP_LIST_PATH", ""));
	
	gSkipStdOut = gConfigTP.mGetNumber("SKIP_STDOUT", 0.0) ? true : false;
	
	gUseManifest = gConfigTP.mGetNumber("USE_MANIFEST", 0.0) ? true : false;
	strcpy(gManifestPath, gConfigTP.mGetString("MANIFEST_PATH", "/iFuntasticMan.txt"));
	
	gGetMD5Date		= gConfigTP.mGetNumber("USE_MD5DATE", 0.0) ? true : false;

	gGetSoundInfo	= gConfigTP.mGetNumber("SOUND_INFO", 0.0) ? true : false;
	gDummyMP3[0]	= 0;
	gDummyMP3Size	= 0;
	
	if (deviceCommand[0])
	{
		if (! strcmp(deviceCommand, "EXIT_RECOVERY")) // 'our' way has problems...
		{	
			sprintf(outStr, "iphonecomm at %s", gPath1);
			my_cout(outStr, kForceOut);
			mExitRecoveryWithIPhonecomm(gPath1);
			my_cout("//OK//", kForceOut);
			goto finish;
		}

		//Call to SERIOUS_HACKERY
		initPrivateFunctions();
		//End SERIOUS_HACKERY

		my_cout("AMDeviceNotificationSubscribe (MD): ", AMDeviceNotificationSubscribe(listDevices_notification, 0, 0, 0, &notif));

		ret = AMRestoreRegisterForDeviceNotifications(
						listDevices_dfu_connect_callback,
						listDevices_recovery_connect_callback,
						dfu_disconnect_callback,
						recovery_disconnect_callback,
						0,
						NULL);

		if (ret != 0) 
		{
			sprintf(outStr, "//ERROR0: AMRestoreRegisterForDeviceNotifications %d//", ret);
			my_cout(outStr, kForceOut);
			goto finish;
		}

		if (! strcmp(deviceCommand, "ENTER_DFU"))
		{
			bSE_deviceCommand("ENTER_RECOVERY", gSerialNo);
			
			if (gWaitForRecoveryIndex != -1)
			{
				if (gDEV[gWaitForRecoveryIndex].sh->shell_mode != SHELL_RECOVERY)
				{
					dispose_shell(gDEV[gWaitForRecoveryIndex].sh);
					MEMSET(&gDEV[gWaitForRecoveryIndex], 0, sizeof(bSEDevice));
					sleep(10); // wait for a recovery connect!
				}
				
				bSE_deviceCommand("COPY_DFU_FILE", gSerialNo);
			}
		}
		else
		{
			bSE_deviceCommand(deviceCommand, gSerialNo);
		}
		
		goto finish;
	}
	
	gVerbose			= gConfigTP.mGetNumber("VERBOSE", 0.0) ? true : false;
	gMultiDevice		= gConfigTP.mGetNumber("MULTI_DEVICE", 0.0) ? true : false;
	gOneShot			= gConfigTP.mGetNumber("ONE_SHOT", 1.0) ? true : false;
	gShowProgress		= gConfigTP.mGetNumber("SHOW_PROGRESS", 0.0) ? true : false;
	gCheckAbort			= gConfigTP.mGetNumber("CHECK_ABORT", 1.0) ? true : false;
	gTimeOut			= gConfigTP.mGetNumber("TIME_OUT", 0.5);
	gRecoveryModesOK	= gConfigTP.mGetNumber("RECOVERY_MODES_OK", 0.0) ? true : false;
	
	//Call to SERIOUS_HACKERY
	initPrivateFunctions(); 
	//End SERIOUS_HACKERY

/*
	if (cli_flags & OPT_RECOVERY)
	{
		D("skipping AMDeviceNotificationSubscribe");
		cout << "iphuc: Waiting for recovery mode callback." << endl;	
	}
*/
	
	my_cout("AMDeviceNotificationSubscribe (MD): ", AMDeviceNotificationSubscribe(notification, 0, 0, 0, &notif));
	
	ret = AMRestoreRegisterForDeviceNotifications(
					dfu_connect_callback,
					recovery_connect_callback,
					dfu_disconnect_callback,
					recovery_disconnect_callback,
					0,
					NULL);
					
	if (ret != 0) 
	{
		my_cout("Problem registering recovery callback: ", ret);
		goto finish;
	}
	
	if (gOneShot)
	{
		// enter a 0.5 second run loop so we don't get stuck even if the phone is not present
		// or not communicating

		my_cout("Time limited CFRunLoop: Waiting for iPhone.");
		res = CFRunLoopRunInMode(kCFRunLoopDefaultMode, gTimeOut, false);
		
		switch(res) // add marker strings to output
		{
			case kCFRunLoopRunFinished:			my_cout("kCFRunLoopRunFinished", kForceOut);		break;
			case kCFRunLoopRunStopped:			my_cout("kCFRunLoopRunStopped", kForceOut);			break;
			case kCFRunLoopRunTimedOut:			my_cout("//STATUS_TIMED_OUT//", kForceOut);			break;
			case kCFRunLoopRunHandledSource:	my_cout("kCFRunLoopRunHandledSource", kForceOut);	break;
		}
	}
	else // use the standard run loop
	{
		my_cout("");
		my_cout("CFRunLoop: Waiting for iPhone.", kForceOut);
		my_cout("");
		CFRunLoopRun();
		my_cout("main: CFRunLoop returned somehow, exiting.  Please report.");
	}

	my_cout("//OK//", kForceOut);

finish:
	
	for (int i = 0; i < kMaxDevices; ++i)
		if (gDEV[i].sh)
		{
			bSE_updateManifest(gDEV[i].sh);
			
			dispose_shell(gDEV[i].sh);
			MEMSET(&gDEV[i], 0, sizeof(bSEDevice));
		}
		
	my_exit();
	return 1;
}

//------------------------------------------------------------------------------------------
#if 0 
#pragma mark -
#endif
//------------------------------------------------------------------------------------------

void bSE_getSerialNo(struct am_device *dev, char *serialNo)
{
	__CFString				*result;
	
	result = AMDeviceCopyValue(dev, 0, CFStringCreateWithCString(NULL, "SerialNumber", kCFStringEncodingASCII));
	
	if (result)
	{
		CFStringGetCString(result, serialNo, sizeof(string256), kCFStringEncodingASCII);
		CFRelease(result);
	}
	else
		serialNo[0] = 0;
		
	return;
}

//------------------------------------------------------------------------------------------

void bSE_getDeviceInfo(struct shell_state *sh)
{
	__CFString				*result;
	struct afc_dictionary	*info;
	string512				outStr, cs;
	
	sprintf(outStr, "//DEVICEINFO%d//", sh->devNo);
	my_cout(outStr, kForceOut);

	result = AMDeviceCopyValue(sh->dev, 0, CFStringCreateWithCString(NULL, "SerialNumber", kCFStringEncodingASCII));
	if (result)
	{
		CFStringGetCString(result, cs, sizeof(cs), kCFStringEncodingASCII);
		CFRelease(result);
		sprintf(outStr, "//SerialNumber%d: %s//", sh->devNo, cs);
		my_cout(outStr, kForceOut);
	}
	
	result = AMDeviceCopyValue(sh->dev, 0, CFStringCreateWithCString(NULL, "ProductVersion", kCFStringEncodingASCII));
	if (result)
	{
		CFStringGetCString(result, cs, sizeof(cs), kCFStringEncodingASCII);
		CFRelease(result);
		sprintf(outStr, "//ProductVersion%d: %s//", sh->devNo, cs);
		my_cout(outStr, kForceOut);
	}

	result = AMDeviceCopyValue(sh->dev, 0, CFStringCreateWithCString(NULL, "FirmwareVersion", kCFStringEncodingASCII));
	if (result)
	{
		CFStringGetCString(result, cs, sizeof(cs), kCFStringEncodingASCII);
		CFRelease(result);
		sprintf(outStr, "//FirmwareVersion%d: %s//", sh->devNo, cs);
		my_cout(outStr, kForceOut);
	}

	result = AMDeviceCopyValue(sh->dev, 0, CFStringCreateWithCString(NULL, "BuildVersion", kCFStringEncodingASCII));
	if (result)
	{
		CFStringGetCString(result, cs, sizeof(cs), kCFStringEncodingASCII);
		CFRelease(result);
		sprintf(outStr, "//BuildVersion%d: %s//", sh->devNo, cs);
		my_cout(outStr, kForceOut);
	}

	result = AMDeviceCopyValue(sh->dev, 0, CFStringCreateWithCString(NULL, "BasebandVersion", kCFStringEncodingASCII));
	if (result)
	{
		CFStringGetCString(result, cs, sizeof(cs), kCFStringEncodingASCII);
		CFRelease(result);
		sprintf(outStr, "//BasebandVersion%d: %s//", sh->devNo, cs);
		my_cout(outStr, kForceOut);
	}

	result = AMDeviceCopyValue(sh->dev, 0, CFStringCreateWithCString(NULL, "ActivationState", kCFStringEncodingASCII));
	if (result)
	{
		CFStringGetCString(result, cs, sizeof(cs), kCFStringEncodingASCII);
		CFRelease(result);
		sprintf(outStr, "//ActivationState%d: %s//", sh->devNo, cs);
		my_cout(outStr, kForceOut);
	}
	
	if (! AFCDeviceInfoOpen(sh->conn, &info))
	{
		char *key = NULL, *val = NULL;
		
		AFCKeyValueRead(info, &key, &val);
		while (key && val) 
		{
			sprintf(outStr,"//%s%d: %s//", key, sh->devNo, val);
			my_cout(outStr, kForceOut);
			
			AFCKeyValueRead(info, &key, &val);
		}
		
		AFCKeyValueClose(info);
	}
	
	//---

	if (! bSE_startAFC2(sh))
	{
		sprintf(outStr,"//AFC2ServiceOK%d//", sh->devNo);
		my_cout(outStr, kForceOut);
		
		// check for iPod software update
		if (! AFCFileInfoOpen(sh->conn, "/private/var/mobile/Media/Nikita/nikita.zip", &info))
		{
			AFCKeyValueClose(info);
			
			sprintf(outStr, "//NikitaPresent%d//", sh->devNo);
			my_cout(outStr, kForceOut);
		}
	
		// throw in the deSadunization check while we're at it
		if (! AFCFileInfoOpen(sh->conn, "/bin/_bSE_alert", &info))
		{
			AFCKeyValueClose(info);
			
			sprintf(outStr, "//NeedToDesadunize%d//", sh->devNo);
			my_cout(outStr, kForceOut);
		}
	
		// iFunVersion check
		{
			UUTextParser	txtTP;
			string512		path = "/private/var/mobile/Media/bSE/iFunVersion.txt";
			
			if (! bSE_readFile(sh, path, &txtTP))
				if (txtTP.fText)
				{
					sprintf(outStr, "//iFunVersion%d: %s//", sh->devNo, txtTP.fText);
					my_cout(outStr, kForceOut);
				}
		}
	
		// get the language setting
		{
			UUTextParser	dataTP, plistTP, TP;
			string512		path = "/private/var/mobile/Library/Preferences/.GlobalPreferences.plist";
			string1K		line;
			
			if (! bSE_readFile(sh, path, &dataTP))
			{
				mConvertPlist(dataTP.fText, dataTP.fTextSize, &plistTP);
				if (! plistTP.mGetTextBetween("<key>AppleLanguages", "</array>", &TP)) // extract the whole language array
					if (! TP.mGetStringBetween("<string>", "</string>", line, sizeof(line)-1)) // the first entry is the main one
					{
						sprintf(outStr, "//LanguageCode%d: %s//", sh->devNo, line);
						my_cout(outStr, kForceOut);
					}
			}
		}

		// get the IP address
		// oops, the file may be in binary format and not readable here... so convert it!
		{
			UUTextParser	dataTP, plistTP;
			string512		path = "/Library/Preferences/SystemConfiguration/com.apple.network.identification.plist";
			string1K		cs, line;
			char			*pC;
			string256		addr, dateStr;
			int				y,m,d,hh,mm,ss;
			int				newestIndex, count;
			double			newestDate, td;
			
			if (! bSE_readFile(sh, path, &dataTP))
			{
				mConvertPlist(dataTP.fText, dataTP.fTextSize, &plistTP);
				
				// pick the *last* one?
				// we may have to look for the actual signature...
				// look for time stamps, then pick the address associated with the newest one
				
				newestIndex = -1;
				newestDate = 0.0;
				count = 0;
				
				plistTP.mSetPos(0);
				while(1)
				{
					if (! plistTP.mGetStringBetween("<key>Timestamp</key>", "</date>", line, sizeof(line)-1))
					{
						strcpy(cs, "<date>");
						pC = strstr(line, cs);
						if (pC)
						{
							pC += strlen(cs);
							strcpy(dateStr, pC); // e.g. 2008-04-13T11:34:13Z
						
							// convert into a double date
							sscanf(dateStr, "%04d-%02d-%02dT%02d:%02d:%02dZ", &y, &m, &d, &hh, &mm, &ss);
							
							td = y; td *= 100;
							td += m; td *= 100;
							td += d; td *= 100;
							td += hh; td *= 100;
							td += mm; td *= 100;
							td += ss;
							
							if (td > newestDate)
								newestIndex = count;
							
							++count;
						}
					}
					else
						break;
				}
				
				addr[0] = 0;
				count = 0;
				plistTP.mSetPos(0);
				while(1)
				{
					if (! plistTP.mGetStringBetween("<key>IPv4", "</string>", line, sizeof(line)-1))
					{
						strcpy(cs, "<string>");
						pC = strstr(line, cs);
						if (pC)
						{
							pC += strlen(cs);
							strcpy(addr, pC);
							
							if (count == newestIndex)
								break;
								
							++count;
						}
					}
					else
						break;
				}
				
				if (addr[0])
				{
					sprintf(outStr, "//IPAddress%d: %s//", sh->devNo, addr);
					my_cout(outStr, kForceOut);
				}
			}
		}

		// get the user-given device name
		{
			UUTextParser	dataTP, plistTP;
			string512		path = "/Library/Preferences/SystemConfiguration/preferences.plist";
			string1K		cs, line;
			char			*pC;
			
			if (! bSE_readFile(sh, path, &dataTP))
			{
				mConvertPlist(dataTP.fText, dataTP.fTextSize, &plistTP);
				
				//if (! plistTP.mGetStringBetween("<key>ComputerName</key>", "</string>", line, sizeof(line)-1))
				// use 'HostName' it seems to not use funky encoding!
				if (! plistTP.mGetStringBetween("<key>HostName</key>", "</string>", line, sizeof(line)-1))
				{
					strcpy(cs, "<string>");
					pC = strstr(line, cs);
					if (pC)
					{
						pC += strlen(cs);
						sprintf(outStr, "//CustomName%d: %s//", sh->devNo, pC);
						my_cout(outStr, kForceOut);
					}
				}
			}
		}

		// get the installer version
		{
			UUTextParser	dataTP, plistTP;
			string512		path = "/Applications/Installer.app/Info.plist";
			string1K		cs, line;
			char			*pC;
			
			if (! bSE_readFile(sh, path, &dataTP))
			{
				mConvertPlist(dataTP.fText, dataTP.fTextSize, &plistTP);
				if (! plistTP.mGetStringBetween("<key>CFBundleVersion</key>", "</string>", line, sizeof(line)-1))
				{
					strcpy(cs, "<string>");
					pC = strstr(line, cs);
					if (pC)
					{
						pC += strlen(cs);
						sprintf(outStr, "//InstallerVersion%d: %s//", sh->devNo, pC);
						my_cout(outStr, kForceOut);
					}
				}
			}
		}

		// check if ssh is present
		{
			if (! AFCFileInfoOpen(sh->conn, "/usr/bin/ssh", &info))
			{
				AFCKeyValueClose(info);
				
				sprintf(outStr, "//sshPresent%d//", sh->devNo);
				my_cout(outStr, kForceOut);
			}
		}
		
		// check script execution
		{
			UUTextParser	scriptTP;
			string512		daemonPath	= "/usr/sbin/_bSE_script6d";
			string512		triggerPath	= "/usr/sbin/_bSE_CHK6.txt";
			string512		okPath		= "/usr/sbin/_bSE_OK6.txt";
			
			// if the daemon exists
			if (! AFCFileInfoOpen(sh->conn, daemonPath, &info))
			{
				AFCKeyValueClose(info);
				
				// trigger the 'check' daemon
				
				AFCRemovePath(sh->conn, okPath);
				if (! bSE_writeFile(sh, triggerPath, "1", 1))
				{
					int	n = 30;
					
					while(1) // wait for the 'OK'
					{
						if (--n > 0)
							usleep(100000); // a tenth of a second
						else
							break;

						if (! AFCFileInfoOpen(sh->conn, okPath, &info))
						{
							AFCKeyValueClose(info);

							sprintf(outStr, "//ScriptExecutionOK%d//", sh->devNo);
							my_cout(outStr, kForceOut);
							break;
						}
					}
				}
			}
		}
	}
	
	sprintf(outStr, "//END_DEVICEINFO%d//\n", sh->devNo);
	my_cout(outStr, kForceOut);
	return;
}

//------------------------------------------------------------------------------------------

int bSE_writeFile(struct shell_state *sh, char *path, char *data, int dataSize)
{
	int				error = 1;
	afc_file_ref	ref = 0;
	afc_error_t		ret;
	
	if (! path[0])
		goto finish;
		
	ret = AFCFileRefOpen(sh->conn, path, 3, &ref);
	if (ret)
		goto finish;
	
	if ((data)&&(dataSize > 0))
	{
		ret = AFCFileRefWrite(sh->conn, ref, data, dataSize);
		if (ret)
			goto finish;
	}
	
	error = 0;
	
finish:

	if (ref)
		AFCFileRefClose(sh->conn, ref);
		
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_readFile(struct shell_state *sh, char *path, UUTextParser *fileTP)
{
	int				error = 1;
	unsigned int	file_size;
	afc_file_ref	ref = 0;
	afc_error_t		ret;
	
	fileTP->mDisposeText();
	
	file_size = get_file_size(sh->conn, path);
	if (! file_size)
		goto finish;
		
	ret = AFCFileRefOpen(sh->conn, path, 1, &ref);
	if (ret)
		goto finish;
	
	fileTP->mSetText(nil, file_size); // allocate
	
	ret = AFCFileRefRead(sh->conn, ref, fileTP->fText, &file_size);
	if (ret)
		goto finish;
	
	fileTP->fTextSize = file_size;
	error = 0;
	
finish:

	if (ref)
		AFCFileRefClose(sh->conn, ref);
		
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_diskDirExists(char *path)
{
	int				exists = 0;
	FSRef			fsRef;
	Boolean			bisDir;
	int				ret;
	
	ret = FSPathMakeRef((UInt8 *) path, &fsRef, &bisDir);
	if (ret) 
	{
		//sprintf(gMsgStr, "//ERROR//FSPathMakeRef_FAILED %d//<%s>//", (long) ret, path);
		//my_cout(gMsgStr, kForceOut);
		goto finish;
	}
	
	exists = (bisDir) ? 1 : 0;
	
finish:
	return exists;
}

//------------------------------------------------------------------------------------------

int bSE_delete(struct shell_state *sh, char *path)
{
	int						error = 1, len;
	struct afc_directory	*dir;
	char					*dirent;
	string512				tpath, absPath;
	
	strcpy(tpath, path);
	len = strlen(tpath);
	if (len > 1)
		if (tpath[ len-1 ] == '/')
			tpath[ len-1 ] = 0;
	
	if (bSE_dirExists(sh, tpath))	// it's a directory
	{
		if (AFCDirectoryOpen(sh->conn, tpath, &dir)) 
		{
			sprintf(gMsgStr, "bSE_delete: could not open directory '%s'.", tpath);
			my_cout(gMsgStr, kForceOutErr);
			goto finish;
		}
		
		// delete the contents
		
		while (1) 
		{
			AFCDirectoryRead(0, dir, &dirent);
			if (! dirent) 
			{
				error = 0; // done
				break;
			}
			
			if ((strcmp(dirent, "."))
				&&(strcmp(dirent, ".."))
			)
			{
				if (! strcmp(tpath, "/"))	sprintf(absPath, "/%s", dirent); //!
				else						sprintf(absPath, "%s/%s", tpath, dirent); //!
				
				if (bSE_delete(sh, absPath))
					break; // error
			}
		}
		
		AFCDirectoryClose(0, dir);

		error = bSE_deleteFile(sh, tpath, true);
	}
	else
	{
		error = bSE_deleteFile(sh, tpath, false);
	}
	
finish:
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_deleteFile(struct shell_state *sh, char *path, bool isDir)
{
	int						error = 1, len;
	afc_error_t				ret;
	string512				tpath;
	
	strcpy(tpath, path);
	len = strlen(tpath);
	if (tpath[ len-1 ] != '/')
		strcat(tpath, "/");
		
	ret = AFCRemovePath(sh->conn, path);
	switch(ret)
	{
		case 0:
		{
			if (bSE_appChange(tpath))
				gAppChange = true;

			bSE_pushManifestItem(sh, kbSE_manifest_deleted, tpath, nil);
		
			sprintf(gMsgStr, "//DELETED//<%s>//", tpath);
			my_cout(gMsgStr, kForceOut);
			error = 0;
		}
		break;
		
		case 1:
		default:
			sprintf(gMsgStr, "bSE_deleteFile: could not delete '%s'.", tpath);
			my_cout(gMsgStr, ret);
		break;
	}
	
finish:
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_deletePlaceholders(struct shell_state *sh)
{
	int			error = 1, terr = 0;
	string512	path;
	
	// delete all iFun??.app folders in /Applications/
	
	for (int i = 0; i < 100; ++i)
	{
		sprintf(path, "/Applications/iFun%02d.app", i);
		if (bSE_dirExists(sh, path))
			terr += bSE_delete(sh, path);
	}
	
	if (! terr)
		error = 0;
		
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_deleteAll_bSE(struct shell_state *sh, char *path)
{
	int						error = 1, len;
	struct afc_directory	*dir;
	char					*dirent;
	string512				tpath, absPath;

	// delete all _bSE_... files in 'path'
	
	strcpy(tpath, path);
	len = strlen(tpath);
	if (len > 1)
		if (tpath[ len-1 ] == '/')
			tpath[ len-1 ] = 0;
	
	if (bSE_dirExists(sh, tpath)) // it's a directory
	{
		if (AFCDirectoryOpen(sh->conn, tpath, &dir)) 
		{
			sprintf(gMsgStr, "bSE_deleteAll_bSE: could not open directory '%s'.", tpath);
			my_cout(gMsgStr, kForceOutErr);
			goto finish;
		}
		
		// delete the contents
		
		while (1) 
		{
			AFCDirectoryRead(0, dir, &dirent);
			if (! dirent) 
			{
				error = 0; // done
				break;
			}
			
			if (! strncmp(dirent, "_bSE_", 5))
			{
				if (! strcmp(tpath, "/"))	sprintf(absPath, "/%s", dirent); //!
				else						sprintf(absPath, "%s/%s", tpath, dirent); //!
				
				if (bSE_deleteFile(sh, absPath, false))
					break; // error
			}
		}
		
		AFCDirectoryClose(0, dir);
	}

finish:		
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_activateSonOf(struct shell_state *sh)
{
	int						error = 1;
	UUTextParser			dataTP;
	
	// copy executable from Preferences.app to iFuntastic.app
	
	if ((bSE_dirExists(sh, "/Applications/iFuntastic.app"))
		&&(bSE_dirExists(sh, "/Applications/Preferences.app"))
	)
	{
		if (! bSE_readFile(sh, "/Applications/Preferences.app/Preferences", &dataTP))
			if (! bSE_writeFile(sh, "/Applications/iFuntastic.app/iFuntasticX", dataTP.fText, dataTP.fTextSize))
				error = 0;
	}

finish:		
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_startAFC2(struct shell_state *sh)
{
	int		error = 1, ret;

	if (sh->shell_mode == SHELL_NORMAL)
		AFCConnectionClose(sh->conn);
		
	// start AFC service
	ret = AMDeviceStartService(sh->dev, AMSVC_AFC2, &(sh->afch), NULL);
	if (! ret)
	{
		// open an AFC Connection
		ret = AFCConnectionOpen(sh->afch, 0, &(sh->conn));
		if (! ret)
		{
			error = 0;
		}
		else
		{
			sprintf(gMsgStr, "//STATUS_AFC2_OPEN_FAILED//ERROR: %ld//", (long) ret);
			my_cout(gMsgStr, kForceOut);
		}
	}
	else
	{
		sprintf(gMsgStr, "//STATUS_AFC2_START_FAILED//ERROR: %ld//", (long) ret);
		my_cout(gMsgStr, kForceOut);
	}
		
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_openConnection(bSEDevice *bSEDev, char *forSerialNo, bool useAfc2)
{
	int					error = 1, ret;
	string256			serialNo;
	struct shell_state	*sh;
	
	sh = bSEDev->sh;
	
	if (! sh)
		goto finish;

	if (sh->shell_mode != SHELL_NORMAL)
		goto finish;
	
	if (! bSEDev->didPair)
	{
		AMDeviceIsPaired(sh->dev);
		ret = AMDeviceValidatePairing(sh->dev);
		if (ret) 
		{
			sprintf(gMsgStr, "//COMM_ERROR//STATUS_PAIRING_FAILED//%d//ERROR: %ld//", sh->devNo, (long) ret);
			my_cout(gMsgStr, kForceOut);
			goto finish;
		}
		
		bSEDev->didPair = true;
	}
	
	if (! bSEDev->didStartSession)
	{
		ret = AMDeviceStartSession(sh->dev);
		if (ret) 
		{
			sprintf(gMsgStr, "//COMM_ERROR//STATUS_SESSION_START_FAILED//%d//ERROR: %ld//", sh->devNo, (long) ret);
			my_cout(gMsgStr, kForceOut);
			goto finish;
		}
		
		bSEDev->didStartSession = true;
	}
	
	if (forSerialNo)
		if (forSerialNo[0])
		{
			bSE_getSerialNo(sh->dev, serialNo);
			if (strcmp(forSerialNo, serialNo))
				goto finish;
		}

	if (useAfc2)
	{
		if (bSEDev->didStartAfc)
		{
			if (bSEDev->didOpenConn)
			{
				AFCConnectionClose(sh->conn);
				bSEDev->didOpenConn = false;
			}
			
			bSEDev->didStartAfc = false; //?
		}
		
		if (! bSEDev->didStartAfc2)
		{
			ret = AMDeviceStartService(sh->dev, AMSVC_AFC2, &(sh->afch), NULL);	
			if (ret) 
			{
				sprintf(gMsgStr, "//COMM_ERROR//STATUS_AFC2_START_FAILED//%d//ERROR: %ld//", sh->devNo, (long) ret);
				my_cout(gMsgStr, kForceOut);
				goto finish;
			}
			
			bSEDev->didStartAfc2 = true;
		}
		
		//AMDeviceStopSession(sh->dev); nope!
		
		if (! bSEDev->didOpenConn)
		{
			ret = AFCConnectionOpen(sh->afch, 0, &(sh->conn));
			if (ret) 
			{
				sprintf(gMsgStr, "//COMM_ERROR//STATUS_AFC2_OPEN_FAILED//%d//ERROR: %ld//", sh->devNo, (long) ret);
				my_cout(gMsgStr, kForceOut);
				goto finish;
			}
		
			bSEDev->didOpenConn = true;
		}
	}
	else
	{
		if (bSEDev->didStartAfc2)
		{
			if (bSEDev->didOpenConn)
			{
				AFCConnectionClose(sh->conn);
				bSEDev->didOpenConn = false;
			}
			
			bSEDev->didStartAfc2 = false; //?
		}

		if (! bSEDev->didStartAfc)
		{
			ret = AMDeviceStartService(sh->dev, AMSVC_AFC, &(sh->afch), NULL);	
			if (ret) 
			{
				sprintf(gMsgStr, "//COMM_ERROR//STATUS_AFC_START_FAILED//%d//ERROR: %ld//", sh->devNo, (long) ret);
				my_cout(gMsgStr, kForceOut);
				goto finish;
			}
		
			bSEDev->didStartAfc = true;
		}
		
		//AMDeviceStopSession(sh->dev); nope!
		
		if (! bSEDev->didOpenConn)
		{
			ret = AFCConnectionOpen(sh->afch, 0, &(sh->conn));
			if (ret) 
			{
				sprintf(gMsgStr, "//COMM_ERROR//STATUS_AFC_OPEN_FAILED//%d//ERROR: %ld//", sh->devNo, (long) ret);
				my_cout(gMsgStr, kForceOut);
				goto finish;
			}
		
			bSEDev->didOpenConn = true;
		}
	}
	
	error = 0;
	
finish:		
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_dataTransfer(struct shell_state *srcSH, char *srcPath, struct shell_state *dstSH, char *dstPath, int transferFlags)
{
	int						error = 1, ret, len;
	bool					isDir = false;
	string512				tSrcPath, tDstPath;
	
	if ((! srcPath)||(! dstPath))
		goto finish;
	if ((! srcPath[0])||(! dstPath[0]))
		goto finish;

	strcpy(tSrcPath, srcPath);
	len = strlen(tSrcPath);
	if (len > 1)
		if (tSrcPath[ len-1 ] == '/')
			tSrcPath[ len-1 ] = 0;

	strcpy(tDstPath, dstPath);
	len = strlen(tDstPath);
	if (len > 1)
		if (tDstPath[ len-1 ] == '/')
			tDstPath[ len-1 ] = 0;
	
	isDir = bSE_dirExists(srcSH, tSrcPath);
	
	if (isDir)	ret = bSE_copyFolder(srcSH, tSrcPath, dstSH, tDstPath, transferFlags);
	else		ret = bSE_copyFile(srcSH, tSrcPath, dstSH, tDstPath, transferFlags);
	
	if (ret)
		goto finish;
	
	error = 0;
	
finish:
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_dataDelete(struct shell_state *srcSH, char *srcPath)
{
	int						error = 1, ret, len;
	string512				tSrcPath;
	bSEDir					theD;
	bSEFile					theF;
	
	if (! srcPath)
		goto finish;
	if (! srcPath[0])
		goto finish;

	strcpy(tSrcPath, srcPath);
	len = strlen(tSrcPath);
	if (len > 1)
		if (tSrcPath[ len-1 ] == '/')
			tSrcPath[ len-1 ] = 0;
	
	if (bSE_dirExists(srcSH, tSrcPath))	
	{
		bSE_dirSet(srcSH, tSrcPath, &theD);

		ret = bSE_dirMakeEmpty(&theD);
		if (ret)
			goto finish;
		
		ret = bSE_dirDelete(&theD);
		if (ret)
			goto finish;
	}
	else		
	{
		bSE_fileSet(srcSH, tSrcPath, &theF);

		ret = bSE_fileDelete(&theF, false);
		if (ret)
			goto finish;
	}
	
	error = 0;
	
finish:
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_dataRename(struct shell_state *sh, char *path, char *newName)
{
	int						error = 1, ret, len;
	string512				dstPath, tpath;
	
	if (! path)
		goto finish;
	if (! path[0])
		goto finish;
	if (! newName)
		goto finish;
	if (! newName[0])
		goto finish;

	// unfortunately this will have to be a copy operation, 
	// followed by a delete...
	
	mSplitPath(path, tpath, nil);

	len = strlen(tpath);
	if (len >= 1) // since we're adding a slash...
		if (tpath[ len-1 ] == '/')
			tpath[ len-1 ] = 0;
	
	sprintf(dstPath, "%s/%s", tpath, newName);
	
	ret = bSE_dataTransfer(sh, path, sh, dstPath, 0);
	if (ret)
		goto finish;

	error = bSE_dataDelete(sh, path);
	
finish:
	return error;
}

//------------------------------------------------------------------------------------------

unsigned long long bSE_totalDataSize(struct shell_state *sh, char *path, int transferFlags)
{
	int						len;
	string512				tpath;
	bSEDir					theD;
	bSEFile					theF;
	unsigned long long		totalSize = 0L;
	
	if (! path)
		goto finish;
	if (! path[0])
		goto finish;

	strcpy(tpath, path);
	len = strlen(tpath);
	if (len > 1)
		if (tpath[ len-1 ] == '/')
			tpath[ len-1 ] = 0;
	
	if (gSkipTP.fText)
	{
		string512 skipPath;
		
		sprintf(skipPath, "\n%s\n", tpath);
		gSkipTP.mSetPos(0);
		if (gSkipTP.mFindString(skipPath))
			goto finish;
	}
		
	if (bSE_dirExists(sh, tpath))	
	{
		totalSize = kDirDummySize; // return a dummy size, so a directory counts in the progress bar
		
		if (transferFlags & kDT_copyFlat)
			goto finish; // done
		
		bSE_dirSet(sh, tpath, &theD);
		if (! bSE_dirOpen(&theD))
		{
			bSEItem	*pI;
			
			pI = &(*theD.PIL)[0];
			for (int i = 0; i < (int) theD.PIL->size(); ++i, ++pI)
				totalSize += bSE_totalDataSize(sh, pI->path, transferFlags);
			
			bSE_dirClose(&theD);
		}
	}
	else		
	{
		bSE_fileSet(sh, tpath, &theF);
		totalSize = bSE_fileSize(&theF);
	}
	
finish:
	return totalSize;
}

//------------------------------------------------------------------------------------------

unsigned int bSE_totalCount(struct shell_state *sh, char *path)
{
	int						len;
	string512				tpath;
	bSEDir					theD;
	unsigned int			totalCount = 0L;
	
	if (! path)
		goto finish;
	if (! path[0])
		goto finish;

	strcpy(tpath, path);
	len = strlen(tpath);
	if (len > 1)
		if (tpath[ len-1 ] == '/')
			tpath[ len-1 ] = 0;
	
	if (bSE_dirExists(sh, tpath))	
	{
		++totalCount;
		
		bSE_dirSet(sh, tpath, &theD);
		if (! bSE_dirOpen(&theD))
		{
			bSEItem	*pI;
			
			pI = &(*theD.PIL)[0];
			for (int i = 0; i < (int) theD.PIL->size(); ++i, ++pI)
				totalCount += bSE_totalCount(sh, pI->path);
			
			bSE_dirClose(&theD);
		}
	}
	else		
	{
		totalCount = 1;
	}
	
finish:
	return totalCount;
}

//------------------------------------------------------------------------------------------

int bSE_dataInfo(struct shell_state *sh, char *path, std::vector<bSEItem> *PIL)
{
	int						error = 0, len, failed;
	string512				tpath;
	bSEFile					theF;
	bSEItem					tI;
	
	MEMSET(&tI, 0, sizeof(tI));
	
	if (! path)
	{error = 1; goto finish;}
	
	if (! path[0])
	{error = 2; goto finish;}

	strcpy(tI.path, path);
	
	strcpy(tpath, path);
	len = strlen(path);
	if (len > 1)
		if (tpath[ len-1 ] == '/')
		{
			tI.isDir = 1;
			tpath[ len-1 ] = 0;
		}
		
	if (bSE_dirExists(sh, tpath))	
	{
		if (gUseManifest)
			bSE_getManifestInfo(sh, tpath, tI.manifestInfo);
		
		tI.isDir = 1;
	}
	else if (tI.isDir)
	{
		tI.isDir = 2; // not found
	}
	else
	{
		bSE_fileSet(sh, tpath, &theF);
		tI.size = bSE_fileSize(&theF, &failed);
		
		if (failed)
			tI.isDir = 3; // not found
		else 
		{
			if (gUseManifest)
				bSE_getManifestInfo(sh, tpath, tI.manifestInfo);

			if (gGetMD5Date)
				bSE_getMD5Date(&theF, tI.size, tI.md5Date);
			
			if (gGetSoundInfo)
				bSE_getSoundInfo(&theF, &tI);
		}
	}
	
	PIL->push_back(tI);
	
finish:
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_listDirectory(struct shell_state *srcSH, char *srcPath)
{
	int						error = 0, n, len;
	string512				tSrcPath;
	bSEDir					theD;
	bSEFile					theF;
	bSEHdr					hdr;
	bSEItem					*pI;
	
	MEMSET(&hdr, 0, sizeof(bSEHdr));
	MEMSET(&theD, 0, sizeof(bSEDir));
	
	if (! srcPath)
	{error = 1; goto finish;}
	
	if (! srcPath[0])
	{error = 2; goto finish;}

	strcpy(tSrcPath, srcPath);
	len = strlen(tSrcPath);
	if (len > 1)
		if (tSrcPath[ len-1 ] == '/')
			tSrcPath[ len-1 ] = 0;
	
	if (bSE_dirExists(srcSH, tSrcPath))	
	{
		bSE_dirSet(srcSH, tSrcPath, &theD);
		if (bSE_dirOpen(&theD))
		{
			error = 4; 
			goto finish;
		}
		
		// fill in the file sizes
		// and manifest info
		
		n = (int) theD.PIL->size();
		pI = &(*theD.PIL)[0];
		for (int i = 0; i < n; ++i, ++pI)
		{
			if (gSkipTP.fText) // links kill the getting of the file size! so skip them
			{
				string512 skipPath;
				
				sprintf(skipPath, "\n%s\n", pI->path);
				gSkipTP.mSetPos(0);
				if (gSkipTP.mFindString(skipPath))
					continue;
			}
			
			if (gUseManifest)
				bSE_getManifestInfo(srcSH, pI->path, pI->manifestInfo);
				
			if (! pI->isDir)
			{
				bSE_fileSet(srcSH, pI->path, &theF);
				pI->size = bSE_fileSize(&theF);
				
				if (gGetMD5Date)
					bSE_getMD5Date(&theF, pI->size, pI->md5Date);

				if (gGetSoundInfo)
					bSE_getSoundInfo(&theF, pI);
			}
		}
	}
	else
	{
		error = 3; 
		goto finish;
	}

finish:
	
	hdr.result = error;
	bSE_writeBinaryOutput(&hdr, theD.PIL);
	
	bSE_dirClose(&theD);
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_listDirectoryDeep(struct shell_state *srcSH, char *srcPath)
{
	int						error = 0;
	std::vector<bSEItem>	IL;
	bSEHdr					hdr;
	
	MEMSET(&hdr, 0, sizeof(bSEHdr));
	
	bSE_listDirectoryDeep2(srcSH, srcPath, &IL);

finish:
	
	hdr.result = error;
	bSE_writeBinaryOutput(&hdr, &IL);
	
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_listDirectoryDeep2(struct shell_state *srcSH, char *srcPath, std::vector<bSEItem> *PIL)
{
	int						error = 0, n, len;
	string512				tSrcPath;
	bSEDir					theD;
	bSEFile					theF;
	bSEHdr					hdr;
	bSEItem					*pI;
	
	if (! srcPath)
	{error = 1; goto finish;}
	
	if (! srcPath[0])
	{error = 2; goto finish;}

	strcpy(tSrcPath, srcPath);
	len = strlen(tSrcPath);
	if (len > 1)
		if (tSrcPath[ len-1 ] == '/')
			tSrcPath[ len-1 ] = 0;
	
	if (bSE_dirExists(srcSH, tSrcPath))	
	{
		bSE_dirSet(srcSH, tSrcPath, &theD);
		if (bSE_dirOpen(&theD))
		{
			error = 4; 
			goto finish;
		}
		
		// fill in the file sizes
		// and manifest info
		
		n = (int) theD.PIL->size();
		pI = &(*theD.PIL)[0];
		for (int i = 0; i < n; ++i, ++pI)
		{
			if (gSkipTP.fText) // links kill the getting of the file size! so skip them
			{
				string512 skipPath;
				
				sprintf(skipPath, "\n%s\n", pI->path);
				gSkipTP.mSetPos(0);
				if (gSkipTP.mFindString(skipPath))
					continue;
			}
			
			if (gUseManifest)
				bSE_getManifestInfo(srcSH, pI->path, pI->manifestInfo);
				
			if (! pI->isDir)
			{
				bSE_fileSet(srcSH, pI->path, &theF);
				pI->size = bSE_fileSize(&theF);
				
				if (gGetMD5Date)
					bSE_getMD5Date(&theF, pI->size, pI->md5Date);

				if (gGetSoundInfo)
					bSE_getSoundInfo(&theF, pI);
			}
			
			PIL->push_back(*pI);
			
			if (pI->isDir)
			{
				bSE_listDirectoryDeep2(srcSH, pI->path, PIL);
			}
		}
	}
	else
	{
		error = 3; 
		goto finish;
	}

finish:
	bSE_dirClose(&theD);
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_writeBinaryOutput(bSEHdr *hdr, std::vector<bSEItem> *PIL)
{
	int						error = 0;
	FILE					*fp = nil;
	bSEItem					*pI;
	
	fp = fopen(gBinOutputPath, "wb");
	if (! fp)
	{error = 1; goto finish;}
	
	if (PIL)
		hdr->numItems = PIL->size();
	
	fwrite(hdr, 1, sizeof(bSEHdr), fp);
	
	if (PIL)
	{
		pI = &(*PIL)[0];
		for (int i = 0; i < hdr->numItems; ++i, ++pI)
		{
			fwrite(pI, 1, sizeof(bSEItem), fp);
		
		#if 0
			sprintf(gMsgStr, "//ITEM %d//<%s>//", i, pI->path);
			my_cout(gMsgStr, kForceOut);
		#endif
		
		}
	}
	
finish:
	if (fp) fclose(fp);
	return error;
}

//------------------------------------------------------------------------------------------
#if 0 
#pragma mark -
#endif
//------------------------------------------------------------------------------------------

void bSE_dirSet(struct shell_state *sh, char *path, bSEDir *d)
{
	int len;
	
	MEMSET(d, 0, sizeof(bSEDir));
	strcpy(d->path, path);
	d->sh = sh;

	len = strlen(d->path);
	if (len > 1)
		if (d->path[ len-1 ] == '/')
			d->path[ len-1 ] = 0;
	
	return;
}

//------------------------------------------------------------------------------------------

int bSE_dirExists(struct shell_state *sh, char *path)
{
	struct afc_directory	*dir;
	int						exists = 0;
	
	if (sh)
	{
		if (! AFCDirectoryOpen(sh->conn, path, &dir))
		{
			AFCDirectoryClose(0, dir);
			exists = true;
		}
	}
	else
	{
		exists = bSE_diskDirExists(path);
	}
	
	//sprintf(gMsgStr, "//IS_DIRECTORY %s//<%s>//", (exists) ? "YES" : "NO", path);
	//my_cout(gMsgStr, kForceOut);
	
	return exists;
}

//------------------------------------------------------------------------------------------

int bSE_dirCreate(bSEDir *d)
{
	int					error = 1, ret = 0;
	
	if (! d->path[0])
		goto finish;
	
	if (d->sh)
	{
		ret = AFCDirectoryCreate(d->sh->conn, d->path);
		if (ret)
			goto finish;
		
		if (bSE_appChange(d->path))
			gAppChange = true;

		bSE_pushManifestItem(d->sh, kbSE_manifest_added, d->path, nil);
	}
	else
	{
		string512	parentPath, name;
		FSRef		parentRef;
		UniChar		ucName[256];
		int			len;
		
		mSplitPath(d->path, parentPath, name);
		ret = FSPathMakeRef((UInt8 *) parentPath, &parentRef, nil);
		if (ret) 
			goto finish;
		
		len = mGetUnicodeString(name, ucName);
		ret = FSCreateDirectoryUnicode(&parentRef, len, ucName, kFSCatInfoNone, nil, nil, nil, nil);
		if (ret) 
			goto finish;
	}
	
	error = 0;
	
finish:

	if (error) 
	{
		sprintf(gMsgStr, "//ERROR//dirCreate_FAILED %d//<%s>//", (long) ret, d->path);
		my_cout(gMsgStr, kForceOut);
	}
	else
	{
		sprintf(gMsgStr, "//ADDED//<%s>//", d->path);
		my_cout(gMsgStr, kForceOut);
	}
	
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_dirOpen(bSEDir *d)
{
	int					error = 1, ret = 0;
	FSIterator			iterator = nil;
	bSEItem				tI;
	char				*dirent;
	
	if (! d->path[0])
		goto finish;
	
	// list all the items in the directory
	
	d->PIL = new std::vector<bSEItem>;

	if (d->sh)
	{
		ret = AFCDirectoryOpen(d->sh->conn, d->path, &d->dir); 
		if (ret)
			goto finish;

		while(1)
		{
			ret = AFCDirectoryRead(0, d->dir, &dirent);
			if (ret)
			{
				sprintf(gMsgStr, "//ERROR//AFCDirectoryRead_FAILED %d//<%s>//", (long) ret, d->path);
				my_cout(gMsgStr, kForceOut);
				goto finish;
			}

			if (! dirent) 
				break;
		
			if ((! strcmp(dirent, "."))
				||(! strcmp(dirent, ".."))
			)
			continue; // skip these two

			MEMSET(&tI, 0, sizeof(tI));
			
			if (! strcmp(d->path, "/"))	sprintf(tI.path, "/%s", dirent);
			else						sprintf(tI.path, "%s/%s", d->path, dirent); //!
			
			tI.isDir = (bSE_dirExists(d->sh, tI.path)) ? 1 : 0;
			
			d->PIL->push_back( tI );
		}
	}
	else
	{
		FSRef					fRef;
		ItemCount				wantObjects = 1, gotObjects;
		Boolean					containerChanged;
		FSCatalogInfoBitmap		whichInfo = kFSCatInfoNone; // kFSCatInfoGettableInfo?
		FSCatalogInfo			theInfo;
		FSRef					theRef;
		FSSpec					theSpec;
		HFSUniStr255			theName;

		// open an iterator
		// can't iterate and delete at the same time...
		// make a path list to be returned one by one instead
		
		ret = FSPathMakeRef((UInt8 *) d->path, &fRef, nil);
		if (ret) 
			goto finish;

		ret = FSOpenIterator(&fRef, kFSIterateFlat, &iterator);
		if (ret) 
			goto finish;
			
		while(1)
		{
			ret = FSGetCatalogInfoBulk(
				iterator, wantObjects, &gotObjects, &containerChanged, whichInfo,	
				&theInfo, &theRef, &theSpec, &theName
			);

			if (ret == errFSNoMoreItems)
				break;

			if (ret)
			{
				sprintf(gMsgStr, "//ERROR//FSGetCatalogInfoBulk_FAILED %d//<%s>//", (long) ret, d->path);
				my_cout(gMsgStr, kForceOut);
				goto finish;
			}
			
			MEMSET(&tI, 0, sizeof(tI));
			
			ret = FSRefMakePath(&theRef, (UInt8 *) tI.path, sizeof(tI.path)-1);
			if (ret)
			{
				sprintf(gMsgStr, "//ERROR//FSRefMakePath_FAILED %d//<%s>//", (long) ret, d->path);
				my_cout(gMsgStr, kForceOut);
				goto finish;
			}
			
			tI.isDir = bSE_diskDirExists(tI.path);
			
			d->PIL->push_back( tI );
		}
	}
	
	error = 0;

finish:
	
	if (iterator) 
		FSCloseIterator(iterator);
		
	if (error) 
	{
		sprintf(gMsgStr, "//ERROR//dirOpen_FAILED %d//<%s>//", (long) ret, d->path);
		my_cout(gMsgStr, kForceOut);
	}

	return error;
}

//------------------------------------------------------------------------------------------

void bSE_dirClose(bSEDir *d)
{
	if (d->dir)
	{
		AFCDirectoryClose(0, d->dir);
		d->dir = nil;
	}
	
	if (d->PIL)
	{
		delete d->PIL;
		d->PIL = nil;
	}
}

//------------------------------------------------------------------------------------------

int bSE_dirNextPath(bSEDir *d, char *nextAbsPath, char *nextName)
{
	int			error = 1, ret = 0;
	string512	tpath;
	
	nextAbsPath[0] = 0;
	
	if (nextName)
		nextName[0] = 0;
	
	if (d->PIL)
	{
		int			n = (int) d->PIL->size();
		bSEItem		tI;
		
		if (! n)
		{
			error = 0; // done
			goto finish;
		}
		
		tI = (*d->PIL)[n-1];
		
		strcpy(nextAbsPath, tI.path);
		
		if (nextName)
			mSplitPath(nextAbsPath, tpath, nextName);
		
		d->PIL->pop_back();
	}
	else
		goto finish;
		
	error = 0;

finish:

	if (error) 
	{
		sprintf(gMsgStr, "//ERROR//dirNextPath_FAILED %d//<%s>//", (long) ret, d->path);
		my_cout(gMsgStr, kForceOut);
	}

	return error;
}

//------------------------------------------------------------------------------------------

int bSE_dirMakeEmpty(bSEDir *d)
{
	int				error = 1, ret = 0;
	string512		nextPath, nextName;
	bool			isDir;
	bSEDir			subDir;
	bSEFile			subFile;
	
	if (bSE_dirOpen(d))
		goto finish;
	
	while(1)
	{
		if (bSE_dirNextPath(d, nextPath, nextName))
			goto finish;
			
		if (! nextPath[0])
			break;
		
		// should we list all entries and *then* delete them?
		
		isDir = bSE_dirExists(d->sh, nextPath);
		if (isDir)
		{
			bSE_dirSet(d->sh, nextPath, &subDir);
			ret = bSE_dirMakeEmpty(&subDir);
			if (ret)
				goto finish;
			
			ret = bSE_dirDelete(&subDir);
			if (ret)
				goto finish;
		}
		else
		{
			bSE_fileSet(d->sh, nextPath, &subFile);
			
			// ignore the result
			// if the file doesn't exist, then we don't care
			// any other problem: deleting the parent directory will fail (since not empty)
			// (or we'll just end up with a non-empty directory - hm)

			bSE_fileDelete(&subFile, false);
			//if (ret) 
			//	goto finish;
		}

		if (bSE_abort()) // abort processing
		{
			ret = 99;
			goto finish;
		}
	}
	
	error = 0;

finish:

	if (error) 
	{
		sprintf(gMsgStr, "//ERROR//dirMakeEmpty_FAILED %d//<%s>//", (long) ret, d->path);
		my_cout(gMsgStr, kForceOut);
	}
	
	bSE_dirClose(d);
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_dirDelete(bSEDir *d)
{
	int			error = 1, ret = 0;
	bSEFile		subFile;
	
	// assumes the directory is empty! (see above)

	if (! d->path[0])
		goto finish;

	if (d->sh) // same function as for files
	{
		if (bSE_appChange(d->path))
			gAppChange = true;

		bSE_fileSet(d->sh, d->path, &subFile);
		
		ret = bSE_fileDelete(&subFile, true);
		if (ret)
			goto finish;
	}
	else
	{
		FSRef	fsRef;
		
		ret = FSPathMakeRef((UInt8 *) d->path, &fsRef, nil);
		if (ret) 
		{
			sprintf(gMsgStr, "//ERROR//FSPathMakeRef_FAILED %d//<%s>//", (long) ret, d->path);
			my_cout(gMsgStr, kForceOut);
			goto finish;
		}

		ret = FSDeleteObject(&fsRef);
		if (ret) 
		{
			sprintf(gMsgStr, "//ERROR//FSDeleteObject_FAILED %d//<%s>//", (long) ret, d->path);
			my_cout(gMsgStr, kForceOut);
			goto finish;
		}
		
		bSE_updateDataDeleteProgress();
	}

	error = 0;
	
finish:

	if (error) 
	{
		sprintf(gMsgStr, "//ERROR//dirDelete_FAILED %d//<%s>//", (long) ret, d->path);
		my_cout(gMsgStr, kForceOut);
	}
	
	return error;
}

//------------------------------------------------------------------------------------------

int bSE_copyFolder(struct shell_state *srcSH, char *srcPath, struct shell_state *dstSH, char *dstPath, int transferFlags)
{
	int					error = 1, ret;
	bSEDir				srcF, dstF;
	bool				deleteDest = false, isDir;
	string512			nextSrcPath, nextName;
	string512			nextDstPath;
	string256			manifestInfo;
	
	if (gSkipTP.fText)
	{
		string512 skipPath;
		
		sprintf(skipPath, "\n%s\n", srcPath);
		gSkipTP.mSetPos(0);
		if (gSkipTP.mFindString(skipPath))
		{
			sprintf(gMsgStr, "//SKIPPED <%s>//", srcPath);
			my_cout(gMsgStr, kForceOut);
			return 0;
		}
	}
	
	//---
	
	bSE_dirSet(srcSH, srcPath, &srcF);
	bSE_dirSet(dstSH, dstPath, &dstF);
	
	if (gDataTransferPreflight)
	{
		bSEItemDT	tDT;
		
		MEMSET(&tDT, 0, sizeof(tDT));
		tDT.srcSH = srcSH;
		strcpy(tDT.srcPath, srcPath);
		tDT.dstSH = dstSH;
		strcpy(tDT.dstPath, dstPath);
		tDT.srcIsDir = true;
		
		tDT.srcSize = kDirDummySize;
		
		gDataTransferIL.push_back(tDT);
	}
	else
	{
		if (bSE_dirExists(dstF.sh, dstF.path))
		{
			if (! (transferFlags & (kDT_copyMerge | kDT_copyFlat)))
			{
				ret = bSE_dirMakeEmpty(&dstF);
				if (ret) 
					goto finish;
			}
		}
		else
		{
			ret = bSE_dirCreate(&dstF);
			if (ret) 
				goto finish;

			//deleteDest = true; ???
		}
		
		bSE_updateDataTransferProgress(kDirDummySize);
	}
	
	if (! (transferFlags & kDT_copyFlat))
	{
		ret = bSE_dirOpen(&srcF);
		if (ret) 
			goto finish;

		while(1)
		{
			if (bSE_dirNextPath(&srcF, nextSrcPath, nextName))
				goto finish;
				
			if (! nextSrcPath[0])
				break;
			
			if (! strcmp(dstPath, "/"))	sprintf(nextDstPath, "/%s", nextName);
			else						sprintf(nextDstPath, "%s/%s", dstPath, nextName);
			
			isDir = bSE_dirExists(srcF.sh, nextSrcPath);
			
			if (isDir)	ret = bSE_copyFolder(srcSH, nextSrcPath, dstSH, nextDstPath, transferFlags);
			else		ret = bSE_copyFile(srcSH, nextSrcPath, dstSH, nextDstPath, transferFlags);
			
			if (ret) 
				goto finish;

			if (bSE_abort()) // abort processing
			{
				ret = 99;
				goto finish;
			}
		}
	}
	
	if (! gDataTransferPreflight)
	{
		// if the file or folder was copied from 
		// another device, then also copy the manifest info (if any)
		
		if (bSE_getManifestInfo(srcSH, srcPath, manifestInfo))	bSE_pushManifestItem(dstSH, kbSE_manifest_setInfo, manifestInfo, dstPath);
		else													bSE_pushManifestItem(dstSH, kbSE_manifest_updated, srcPath, dstPath);

		sprintf(gMsgStr, "//UPDATED//<%s>//<%s>//", srcPath, dstPath);
		my_cout(gMsgStr, kForceOut);
	}
	
	error = 0;
	
finish:
	
	bSE_dirClose(&srcF);
	bSE_dirClose(&dstF);
	
	if ((error)&&(deleteDest)) // delete the destination folder
	{
		bSE_dirMakeEmpty(&dstF);
		bSE_dirDelete(&dstF);
	}
	
	if (error) 
	{
		sprintf(gMsgStr, "//ERROR//copyFolder_FAILED %d//<%s>//<%s>//", (long) ret, srcPath, dstPath);
		my_cout(gMsgStr, kForceOut);
	}

	return error;
}

//------------------------------------------------------------------------------------------
#if 0 
#pragma mark -
#endif
//------------------------------------------------------------------------------------------

void bSE_fileSet(struct shell_state *sh, char *path, bSEFile *f)
{
	MEMSET(f, 0, sizeof(bSEFile));
	strcpy(f->path, path);
	f->sh = sh;
	return;
}

//------------------------------------------------------------------------------------------

int bSE_fileOpen(bSEFile *f, bool readOnly)
{
	int					error = 1, ret = 0;
	int					mode;
	
	if (! f->path[0])
		goto finish;
		
	if (f->sh)
	{
		#if 0
			// damn! we can't determine if a file is link or not!
			//  and if we *do* open a link, then things get screwed up from then on!
			{
				struct afc_dictionary	*info;
				char					*key = nil, *val = nil;

				sprintf(gMsgStr, "//FILE_INFO//<%s>//", f->path);
				my_cout(gMsgStr, kForceOut);
				
				if (! AFCFileInfoOpen(f->sh->conn, f->path, &info))
				{
					AFCKeyValueRead(info, &key, &val);
					while (key && val) 
					{
						sprintf(gMsgStr, "//KEY %s VALUE %s//", key, val);
						my_cout(gMsgStr, kForceOut);
						AFCKeyValueRead(info, &key, &val);
					}
					
					AFCKeyValueClose(info);
				}
			}
		#endif
		
		#if 0
			
			ret = AFCFileRefOpen(f->sh->conn, f->path, 8, &f->ref);
			sprintf(gMsgStr, "//AFCFileRefOpen Mode 8; ret: %d, ref: %s <%s>//", ret, (f->ref) ? "YES" : "NO", f->path);
			my_cout(gMsgStr, kForceOut);
			if (f->ref)
			{
				AFCFileRefClose(f->sh->conn, f->ref);
				f->ref = nil;
			}
		
		#endif		
		
		//---
		
		if (readOnly)	mode = 1;
		else			mode = 3;

		ret = AFCFileRefOpen(f->sh->conn, f->path, mode, &f->ref);
		if (ret)
		{
			if (f->ref)
			{
				AFCFileRefClose(f->sh->conn, f->ref);
				f->ref = nil;
			}
			
			//error = ret; // so we can examine the result (e.g. treat '11' as possible link)
			//gotta abort - if we ran into a link, nothing works anymore...
			goto finish;
		}
	}
	else
	{
		if (readOnly)	f->fp = fopen(f->path, "r");
		else			f->fp = fopen(f->path, "w");
		
		if (! f->fp)
			goto finish;
	}
	
	error = 0;
	
finish:

	if (error) 
	{
		sprintf(gMsgStr, "//ERROR//fileOpen_FAILED %d//<%s>//", (long) ret, f->path);
		my_cout(gMsgStr, kForceOut);
	}

	return error;
}

//------------------------------------------------------------------------------------------

void bSE_fileClose(bSEFile *f)
{
	if (f->ref)
	{
		if (f->sh)
			AFCFileRefClose(f->sh->conn, f->ref);
		
		f->ref = 0;
	}
	
	if (f->fp)
	{
		fclose(f->fp);
		f->fp = nil;
	}
	
	return;
}

//------------------------------------------------------------------------------------------

int bSE_fileDelete(bSEFile *f, bool isDir)
{
	int	ret = 0;
	
	if (! f->path[0])
		goto finish;

	if (f->sh)
	{
		ret = bSE_deleteFile(f->sh, f->path, true);
		
		if (! ret)
			bSE_updateDataDeleteProgress();
	}
	else
	{
		unlink(f->path);
		ret = 0;
	}

finish:	
	return ret;
}

//------------------------------------------------------------------------------------------

unsigned long bSE_fileSize(bSEFile *f, int *pfailed)
{
	unsigned long			fileSize = 0L;
	FILE					*fp;
	int						error = 0;
	
	if (f->sh)
	{
		//fileSize = get_file_size(f->sh->conn, f->path);
		
		if (f->sh->conn)
		{
			struct afc_dictionary	*info = nil;
			char					*key = nil, *val = nil;
			
			if (AFCFileInfoOpen(f->sh->conn, f->path, &info))
			{
				sprintf(gMsgStr, "//ERROR//fileSize_FAILED//<%s>//", f->path);
				my_cout(gMsgStr, kForceOut);
				error = 1;
			}
			else
			{
				AFCKeyValueRead(info, &key, &val);
				while (key && val) 
				{
					if (! strcmp(key, "st_size"))
					{
						sscanf(val, "%lu", &fileSize);
						break;
					}
					
					AFCKeyValueRead(info, &key, &val);
				}
			}
			
			if (info)
				AFCKeyValueClose(info);
		}
	}
	else
	{
		fp = fopen(f->path, "r");
		if (fp)
		{
			fseek( fp, 0L, SEEK_END );
			fileSize = ftell( fp );
			fclose(fp);
		}
		else
		{
			error = 1;
		}
	}
	
	if (pfailed)
		*pfailed = error;
		
	return fileSize;
}

//------------------------------------------------------------------------------------------

void bSE_getMD5Date(bSEFile *f, unsigned long size, char *md5Date)
{
	char			*s;
	unsigned char	buffer[1024];
	int				ret;
	long			offset, bufSize = 1024L;
	MD5_RD3			MD5;
	
	MEMSET(md5Date, 0, 32L);
	
	ret = bSE_fileOpen(f, true);
	if (ret) 
		goto finish;
	
	if (! size)
		goto finish;
		
	if (size <= bufSize)
	{
		MEMSET(buffer, 0, bufSize);
		
		if (! bSE_fileRead(f, buffer, size))
		{
			MD5.update(buffer, bufSize);
			MD5.finalize();
			s = MD5.hex_digest();
			if (s)
			{
				strcpy(md5Date, s);
				delete s;
			}
		}
	}
	else
	{
		offset = 0;
		MEMSET(buffer, 0, bufSize);
		bSE_fileRead(f, buffer, bufSize);
		offset += bufSize;
		
		// take n samples from the file?
		// for now, just read the first KB (like above)
		
		MD5.update(buffer, bufSize);
		MD5.finalize();
		s = MD5.hex_digest();
		if (s)
		{
			strcpy(md5Date, s);
			delete s;
		}
	}

finish:	
	bSE_fileClose(f);
	return;
}

//------------------------------------------------------------------------------------------

void bSE_getSoundInfo(bSEFile *f, bSEItem *pI)
{
	if (strstr(f->path, ".mp3"))
	{
		bSE_getSoundInfo_MP3(f, pI);
	}
	else if (strstr(f->path, ".m4p"))
	{
		bSE_getSoundInfo_AAC(f, pI);
	}
	else if (strstr(f->path, ".m4a"))
	{
		bSE_getSoundInfo_AAC(f, pI);
	}
	
	return;
}

//------------------------------------------------------------------------------------------

void bSE_getSoundInfo_AAC(bSEFile *f, bSEItem *pI)
{
	unsigned char	buffer[16384], *buf = nil;
	long			bufSize = 2048, bufSize2 = 128L * 1024L;
	int				ret, fileSize;
	string512		path;
	FILE			*fp;
	
	if (! f->sh)
	{
		mGetSoundInfo_AAC(f->path, pI);
		goto finish;
	}
	
	if ((! strstr(f->path, ".m4a"))
		&&(! strstr(f->path, ".m4p"))
	)
	goto finish;
		
	if (! pI->size)
		goto finish;
	
	ret = bSE_fileOpen(f, true);
	if (ret) 
		goto finish;
	
	if (pI->size < bufSize)
		bufSize = pI->size;
			
	MEMSET(buffer, 0, bufSize);
	
	//---
	// read the first 2K; tags are usually/hopefully stored there
	
	if (! bSE_fileRead(f, buffer, bufSize))
	{
		fileSize = bufSize;
		
		strcpy(path, "tmp.m4a");
		unlink(path);
		
		fp = fopen(path, "wb");
		if (! fp)
			goto finish;
			
		fwrite(buffer, 1, fileSize, fp);
		fclose(fp);
		
		if (! mGetSoundInfo_AAC(path, pI))
			goto finish;
	}

	bSE_fileClose(f);
	
	//---
	// not found - may be later in the file
	// read a bigger chunk and try again

	ret = bSE_fileOpen(f, true);
	if (ret) 
		goto finish;

	if (pI->size < bufSize2)
		bufSize2 = pI->size;

	buf = (unsigned char *) CALLOC(bufSize2);
	if (! buf)
	{
		sprintf(gMsgStr, "//ERROR//CALLOC_FAILED//");
		my_cout(gMsgStr, kForceOut);
		goto finish;
	}

	if (! bSE_fileRead(f, buf, bufSize2))
	{
		fileSize = bufSize2;
		
		strcpy(path, "tmp.m4a");
		unlink(path);
		
		fp = fopen(path, "wb");
		if (! fp)
			goto finish;
			
		fwrite(buf, 1, fileSize, fp);
		fclose(fp);
		
		if (! mGetSoundInfo_AAC(path, pI))
			goto finish;
	}
	
finish:	

	if (buf)
		free(buf);

	bSE_fileClose(f);
	return;
}

//------------------------------------------------------------------------------------------

void bSE_getSoundInfo_MP3(bSEFile *f, bSEItem *pI)
{
	unsigned char	buffer[16384], buffer2[1024];
	long			bufSize = 3072L, bufSize2 = 1024L;
	int				ret, fileSize;
	string512		path;
	FILE			*fp;
	unsigned char	*buf = nil;
	
	if (! f->sh)
	{
		mGetSoundInfo_MP3(f->path, pI);
		goto finish;
	}
	
	if (! strstr(f->path, ".mp3"))
		goto finish;
		
	if (! pI->size)
		goto finish;
	
	ret = bSE_fileOpen(f, true);
	if (ret) 
		goto finish;
	
	if (pI->size < bufSize)
		bufSize = pI->size;
			
	MEMSET(buffer, 0, bufSize);
	
	if (! gDummyMP3Size) // load a dummy mp3 file to provide valid MPEG data
	{
		fp = fopen("dummy.mp3", "r");
		if (! fp)
			goto finish;
		
		fseek( fp, 0L, SEEK_END );
		gDummyMP3Size = ftell( fp );
		fseek( fp, 0L, SEEK_SET );
		fread(gDummyMP3, 1, gDummyMP3Size, fp);
		fclose(fp);
	}
	
	//---
	// read the first 3K; ID3 tags are usually stored there
	
	if (! bSE_fileRead(f, buffer, bufSize))
	{
		fileSize = bufSize+2+gDummyMP3Size;
		
		// copy the dummy MPEG data into the buffer
		MEMCPY(&buffer[bufSize+2], gDummyMP3, gDummyMP3Size);
	
		strcpy(path, "tmp.mp3");
		fp = fopen(path, "wb");
		if (! fp)
			goto finish;
			
		fwrite(buffer, 1, fileSize, fp);
		fclose(fp);
		
		if (! mGetSoundInfo_MP3(path, pI))
			goto finish;
	}

	bSE_fileClose(f);
	
	//---
	// no good
	// could also be at the end of the file (especially older version 'TAG' data)
	
	ret = bSE_fileOpen(f, true);
	if (ret) 
		goto finish;
	
	//sprintf(gMsgStr, "//bSE_getSoundInfo_MP3 second try %s//", f->path);
	//my_cout(gMsgStr, kForceOut);

	// no idea about the parameters - assume analogous to fseek
	bSE_fileSeek(f, pI->size - bufSize2, 0L); 
	
	if (! bSE_fileRead(f, buffer2, bufSize2))
	{
		//sprintf(gMsgStr, "//bSE_getSoundInfo_MP3 read tail %s//", f->path);
		//my_cout(gMsgStr, kForceOut);

		// look for 'TAG'
		for (int i = 0; i < bufSize2-3; ++i)
			if (! strncmp((char *) &buffer2[i], "TAG", 3))
			{
				//sprintf(gMsgStr, "//bSE_getSoundInfo_MP3 found TAG %s//", f->path);
				//my_cout(gMsgStr, kForceOut);

				fileSize = gDummyMP3Size+bufSize2-i;
		
				// copy the dummy MPEG data into the buffer
				MEMCPY(buffer, gDummyMP3, gDummyMP3Size);
				// append the TAG data
				MEMCPY(&buffer[gDummyMP3Size], &buffer2[i], bufSize2-i);
			
				// give it a shot

				strcpy(path, "tmp.mp3");
				fp = fopen(path, "wb");
				if (! fp)
					goto finish;
				
				fwrite(buffer, 1, fileSize, fp);
				fclose(fp);
				
				if (! mGetSoundInfo_MP3(path, pI))
					goto finish;
			}
	}

	bSE_fileClose(f);
	
	//---
	// still no good
	// read - nay, COPY the whole frigging file?
	// it'll take forever
	
/* yep, this is way too slow...

	ret = bSE_fileOpen(f, true);
	if (ret) 
		goto finish;

	buf = (unsigned char *) CALLOC(pI->size);
	if (! buf)
	{
		sprintf(gMsgStr, "//ERROR//CALLOC_FAILED//");
		my_cout(gMsgStr, kForceOut);
		goto finish;
	}

	if (! bSE_fileRead(f, buf, pI->size))
	{
		strcpy(path, "tmp.mp3");
		fp = fopen(path, "wb");
		if (! fp)
			goto finish;
		
		fwrite(buf, 1, pI->size, fp);
		fclose(fp);
		
		if (! mGetSoundInfo_MP3(path, pI))
			goto finish;
	}

*/
	
finish:	

	if (buf)
		free(buf);

	bSE_fileClose(f);
	return;
}

//------------------------------------------------------------------------------------------

int bSE_fileRead(bSEFile *f, unsigned char *buf, unsigned int size)
{
	int					error = 1, ret = 0;
	
	if (size)
	{
		if (! buf)
			goto finish;
	
		if (f->sh)
		{
			if (! f->ref)
				goto finish;
				
			ret = AFCFileRefRead(f->sh->conn, f->ref, buf, &size);
			if (ret) 
				goto finish;
		}
		else if (f->fp)
		{
			fread(buf, size, 1, f->fp);
		}
		else
			goto finish;
	}
	
	error = 0;
	
finish:	

	if (error) 
	{
		sprintf(gMsgStr, "//ERROR//fileRead_FAILED %d//<%s>//%lu//", (long) ret, f->path, (unsigned long) size);
		my_cout(gMsgStr, kForceOut);
	}

	return error;
}

//------------------------------------------------------------------------------------------

int bSE_fileWrite(bSEFile *f, unsigned char *buf, unsigned int size)
{
	int					error = 1, ret = 0;
	
	if (size)
	{
		if (! buf)
			goto finish;
	
		if (f->sh)
		{
			if (! f->ref)
				goto finish;
				
			ret = AFCFileRefWrite(f->sh->conn, f->ref, buf, size);
			if (ret) 
				goto finish;
		}
		else if (f->fp)
		{
			fwrite(buf, size, 1, f->fp);
		}
		else
			goto finish;
	}
	
	error = 0;
	
finish:	

	if (error) 
	{
		sprintf(gMsgStr, "//ERROR//fileWrite_FAILED %d//<%s>//%lu//", (long) ret, f->path, (unsigned long) size);
		my_cout(gMsgStr, kForceOut);
	}

	return error;
}

//------------------------------------------------------------------------------------------

int bSE_fileSeek(bSEFile *f, unsigned long long offset1, unsigned long long offset2)
{
	int					error = 1, ret = 0;
	unsigned char		*buf = nil;
	
	if (f->sh)
	{
		if (! f->ref)
			goto finish;
	
		ret = AFCFileRefSeek(f->sh->conn, f->ref, offset1, offset2);
		if (ret) 
			goto finish;
	}
	else if (f->fp)
	{
		fseek(f->fp, offset1, SEEK_SET);
	}
	else
		goto finish;
	
	error = 0;
	
finish:	

	if (error) 
	{
		sprintf(gMsgStr, "//ERROR//fileSeek_FAILED %d//<%s>//%lu//%lu//", (long) ret, f->path, (unsigned long) offset1, (unsigned long) offset2);
		my_cout(gMsgStr, kForceOut);
	}

	if (buf)
		free(buf);

	return error;
}

//------------------------------------------------------------------------------------------

int bSE_copyFile(struct shell_state *srcSH, char *srcPath, struct shell_state *dstSH, char *dstPath, int transferFlags)
{
	int					error = 1, ret, count;
	bSEFile				srcF, dstF;
	unsigned int		srcDataSize;
	unsigned int		chunk;
	unsigned int		maxChunk = 8*1024*1024;
	unsigned char		*buf = nil;
	unsigned int		copiedBytes; // make this long long ?
	bool				deleteDest;
	string256			manifestInfo;
	
	if (gSkipTP.fText)
	{
		string512 skipPath;
		
		sprintf(skipPath, "\n%s\n", srcPath);
		gSkipTP.mSetPos(0);
		if (gSkipTP.mFindString(skipPath))
		{
			sprintf(gMsgStr, "//SKIPPED <%s>//", srcPath);
			my_cout(gMsgStr, kForceOut);
			return 0;
		}
	}

	if (gDataTransferPreflight)
	{
		bSEItemDT	tDT;
		
		MEMSET(&tDT, 0, sizeof(tDT));
		tDT.srcSH = srcSH;
		strcpy(tDT.srcPath, srcPath);
		tDT.dstSH = dstSH;
		strcpy(tDT.dstPath, dstPath);

		bSE_fileSet(tDT.srcSH, tDT.srcPath, &srcF);
		tDT.srcSize = bSE_fileSize(&srcF);
		
		gDataTransferIL.push_back(tDT);
		return 0;
	}
	
	//---
	
	bSE_fileSet(srcSH, srcPath, &srcF);
	bSE_fileSet(dstSH, dstPath, &dstF);
	
	srcDataSize = bSE_fileSize(&srcF);
	
	if (srcDataSize)
	{
		ret = bSE_fileOpen(&srcF, true);
		if (ret) 
			goto finish;
	}
	
	ret = bSE_fileOpen(&dstF, false); // this creates the file, even if empty
	if (ret) 
		goto finish;
		
	if (! srcDataSize)
		goto done; // not an error!
	
	deleteDest = true;
	
	if (srcDataSize > maxChunk)
	{
		sprintf(gMsgStr, "//COPYING %d//", srcDataSize);
		my_cout(gMsgStr, kForceOut);

		buf = (unsigned char *) CALLOC(maxChunk);
		if (! buf)
		{
			sprintf(gMsgStr, "//ERROR//CALLOC_FAILED//");
			my_cout(gMsgStr, kForceOut);
			goto finish;
		}
		
		count = 0;
		copiedBytes = 0L;
		while(copiedBytes < srcDataSize)
		{
			chunk = maxChunk;

			if (copiedBytes+chunk >= srcDataSize)
				chunk = srcDataSize-copiedBytes;

			if (bSE_fileRead(&srcF, buf, chunk))
				goto finish;
			
			if (bSE_fileWrite(&dstF, buf, chunk))
				goto finish;
			
			copiedBytes += chunk;
			bSE_updateDataTransferProgress(chunk);
			
			//if ((count % 5) == 0)	// limit the check frequency
				if (bSE_abort())	// abort processing
				{
					ret = 99;
					break;
				}
			
			++count;
		}
	}
	else
	{
		buf = (unsigned char *) CALLOC(srcDataSize);
		if (! buf)
		{
			sprintf(gMsgStr, "//ERROR//CALLOC_FAILED//");
			my_cout(gMsgStr, kForceOut);
			goto finish;
		}
		
		if (bSE_fileRead(&srcF, buf, srcDataSize))
			goto finish;
		
		if ((transferFlags & kDT_convertPlists)
			&&(! dstSH)
			&&((strstr(dstF.path, ".plist"))
				||(strstr(dstF.path, ".strings"))
			)
		)
		{
			UUTextParser TP;
			
			ret = mConvertPlist((char *) buf, srcDataSize, &TP);
			if (! ret)
			{
				if (TP.mWriteTextToFile(dstF.path))
					goto finish;
			}
			else
			{
				sprintf(gMsgStr, "//WARNING//conversion_FAILED %d//<%s>", ret, dstPath);
				my_cout(gMsgStr, kForceOut);

				if (bSE_fileWrite(&dstF, buf, srcDataSize))
					goto finish;
			}
		}
		else
		{
			if (bSE_fileWrite(&dstF, buf, srcDataSize))
				goto finish;
			
			//---
			
			bSE_fileClose(&dstF);

			if ((transferFlags & kDT_convertPng)
				&&(! dstSH)
				&&((strstr(dstF.path, ".png"))
					||(strstr(dstF.path, ".PNG"))
				)
			)
			{
				// convert the file in place
				mConvertPng(dstF.path);
			}
		}
		
		bSE_updateDataTransferProgress(srcDataSize);
	}

done:
	
	// if the file or folder was copied from 
	// another device, then also copy the manifest info (if any)
	
	if (bSE_getManifestInfo(srcSH, srcPath, manifestInfo))	bSE_pushManifestItem(dstSH, kbSE_manifest_setInfo, manifestInfo, dstPath);
	else													bSE_pushManifestItem(dstSH, kbSE_manifest_updated, srcPath, dstPath);
	
	sprintf(gMsgStr, "//UPDATED//<%s>//<%s>//%lu//", srcPath, dstPath, srcDataSize);
	my_cout(gMsgStr, kForceOut);
	
	error = 0;
	
finish:
	
	if (buf)
		free(buf);
	
	bSE_fileClose(&srcF);
	bSE_fileClose(&dstF);
	
	if ((error)&&(deleteDest)) // delete the destination file
		bSE_fileDelete(&dstF, false);
	
	if (error) 
	{
		sprintf(gMsgStr, "//ERROR//copyFile_FAILED %d//<%s>//<%s>//", (long) ret, srcPath, dstPath);
		my_cout(gMsgStr, kForceOut);
	}

	return error;
}

//------------------------------------------------------------------------------------------
#if 0 
#pragma mark -
#endif
//------------------------------------------------------------------------------------------

void bSE_pushManifestItem(struct shell_state *sh, int action, const char *path1, const char *path2)
{
	bSEManifestItem		tMF;
	
	if (! sh)
		goto finish;
	
	if (sh->shell_mode != SHELL_NORMAL)
		goto finish;
	
	if (! gUseManifest)
		goto finish;
	
	if (! sh->MF)
		sh->MF = new std::vector<bSEManifestItem>;
		
	MEMSET(&tMF, 0, sizeof(bSEManifestItem));
	
	tMF.action = action;
	
	if (path1)
		strcpy(tMF.path1, path1);

	if (path2)
		strcpy(tMF.path2, path2);
	
	sh->MF->push_back(tMF);

finish:
	return;
}

//------------------------------------------------------------------------------------------

void bSE_loadManifest(struct shell_state *sh)
{
	//sprintf(gMsgStr, "loading manifest %s", gManifestPath);
	//my_cout(gMsgStr, kForceOut);
	
	if (! sh->manifestTP)
		sh->manifestTP = new UUTextParser;
		
	bSE_readFile(sh, gManifestPath, sh->manifestTP);
	
	//sprintf(gMsgStr, "manifest size %d", sh->manifestTP->fTextSize);
	//my_cout(gMsgStr, kForceOut);
	return;
}

//------------------------------------------------------------------------------------------

bool bSE_getManifestInfo(struct shell_state *sh, char *path, char *manifestInfo)
{
	bool		found = false;
	string256	start, end;
	string512	tpath;
	int			len;
	
	manifestInfo[0] = 0;

	if (! sh)
		goto finish;

	if (sh->shell_mode != SHELL_NORMAL)
		goto finish;

	if (! gUseManifest)
		goto finish;

	if (! sh->manifestTP)
		goto finish;

	if (! sh->manifestTP->fTextSize)
		goto finish;

	strcpy(tpath, path);
	len = strlen(tpath);
	if (len > 1)
		if (tpath[ len-1 ] == '/')
			tpath[ len-1 ] = 0;

	sprintf(start, "%s\n= ", tpath);
	sprintf(end, "\n");
	
	sh->manifestTP->mSetPos(0);
	if (! sh->manifestTP->mGetStringBetween(start, end, manifestInfo, sizeof(string256)-1))
		found = true;
	
finish:
	return found;
}
	
//------------------------------------------------------------------------------------------

void bSE_updateManifest(struct shell_state *sh)
{
	bSEManifestItem		*pMF;
	string512			path1, path2, name1, name2;
	bool				doWrite = false;
	string256			manifestInfo;
	
	if (! sh)
		goto finish;

	if (sh->shell_mode != SHELL_NORMAL)
		goto finish;

	if (! gUseManifest)
		goto finish;

	if (! sh->MF)
		goto finish;
	
	bSE_loadManifest(sh); // get the current file!

	if (! sh->manifestTP)
		goto finish;

	// get the manifest file
	// we do operations that produce manifest entries one device at a time
	// this is the path on the *destination* device for transfers
	// or *the* device for deletions and 'normal' (via exec_line) operations 
	
	pMF = &(*sh->MF)[0];
	for (int i = 0; i < (int) sh->MF->size(); ++i, ++pMF)
	{
		switch(pMF->action)
		{
			case kbSE_manifest_added:
				bSE_updateManifestTP(sh->manifestTP, pMF->path1, "+");
				doWrite = true;
			break;
			
			case kbSE_manifest_deleted:
				bSE_updateManifestTP(sh->manifestTP, pMF->path1, nil);
				doWrite = true;
			break;
			
			case kbSE_manifest_updated:
				
				mSplitPath(pMF->path1, path1, name1); // source
				mSplitPath(pMF->path2, path2, name2); // destination
				
				if (! strcmp(name1, name2))	strcpy(manifestInfo, "+");
				else						sprintf(manifestInfo, "(%s)", name1);

				bSE_updateManifestTP(sh->manifestTP, pMF->path2, manifestInfo);

				doWrite = true;
			break;
			
			case kbSE_manifest_setInfo: // 'path1' is the manifest info to enter for path2
				bSE_updateManifestTP(sh->manifestTP, pMF->path2, pMF->path1);
				doWrite = true;
			break;
			
			case kbSE_manifest_copied:
			case kbSE_manifest_backedUp:
			case kbSE_manifest_restored:
			default:
			break;
		}
	}
	
	if (doWrite)
	{
		//sprintf(gMsgStr, "updating manifest; size %d", sh->manifestTP->fTextSize);
		//my_cout(gMsgStr, kForceOut);
		
		bSE_writeFile(sh, gManifestPath, sh->manifestTP->fText, sh->manifestTP->fTextSize);

		//sprintf(gMsgStr, "manifest written");
		//my_cout(gMsgStr, kForceOut);
	}
	
finish:

	if (sh->MF)
		sh->MF->clear();
	
	return;
}

//----------------------------------------------------------------------------------------------

void bSE_updateManifestTP(UUTextParser *TP, const char *phonePath, const char *text)
{
	errcode		err = 0;
	string1K	line, dataStr, findStr, start, end;
	
	if (! TP)
	{err = 1; goto finish;}

	sprintf(start, "%s\n= ", phonePath);
	sprintf(end, "\n");
	
	TP->mSetPos(0);
	if (! TP->mGetStringBetween(start, end, dataStr, sizeof(dataStr)-1))
	{
		// found it
		sprintf(findStr, "%s%s%s", start, dataStr, end);
		
		TP->mSetPos(0);
		TP->mReplaceString(findStr, "");
	}
	
	if (text) // add it
	{
		sprintf(line, "%s%s%s", start, text, end);
		TP->mAddString(line);
	}

finish:
	return;
}

//------------------------------------------------------------------------------------------
#if 0 
#pragma mark -
#endif
//------------------------------------------------------------------------------------------

bool bSE_appChange(char *path)
{
	bool		yes = false;
	string512	check, tpath;
	int			pathLen, l;
	
	if ((! strstr(path, ".app"))
		&&(! strstr(path, ".webclip"))
	)
	goto finish;
	
	strcpy(tpath, path);	
	pathLen = strlen(tpath);

	if (pathLen > 1)
		if (tpath[ pathLen-1 ] == '/')
		{
			tpath[ pathLen-1 ] = 0;
			--pathLen;
		}
		
	strcpy(check, "/Applications/"); l = strlen(check);
	if (! strncmp(&tpath[0], check, l))
	{
		strcpy(check, ".app"); l = strlen(check);
		if (pathLen > l)
			if (! strncmp(&tpath[ pathLen-l ], check, l))
			{
				sprintf(gMsgStr, "//APP//<%s/>", tpath);
				my_cout(gMsgStr, kForceOut);

				yes = true;
				goto finish;
			}
	}
	
	strcpy(check, "/private/var/mobile/Library/WebClips/"); l = strlen(check);
	if (! strncmp(&tpath[0], check, l))
	{
		strcpy(check, ".webclip"); l = strlen(check);
		if (pathLen > l)
			if (! strncmp(&tpath[ pathLen-l ], check, l))
			{
				sprintf(gMsgStr, "//WEBCLIP//<%s/>", tpath);
				my_cout(gMsgStr, kForceOut);

				yes = true;
				goto finish;
			}
	}

finish:
	return yes;
}

//------------------------------------------------------------------------------------------

bool bSE_abort(void)
{
	bool	abort = false;
	FILE	*fp;
	
	if (gCheckAbort)
	{
		fp = fopen(gAbortPath, "r");
		if (fp)
		{
			fclose(fp);
			abort = true;
		}
	}

	return abort;
}

//------------------------------------------------------------------------------------------

void bSE_updateProgress(int numDone, int numTotal)
{
	string256		outStr;
	FILE			*fp = nil;
	
	if (gShowProgress)
	{
		if (numTotal)
		{
			if (numTotal < 0)	sprintf(outStr, "-1");
			else				sprintf(outStr, "%d", (int) ((double) numDone * 100.0 / (double) numTotal + 0.5));
			
			fp = fopen(gProgressPath, "w");
			if (fp) 
			{
				fwrite(outStr, 1, strlen(outStr), fp);
				fclose(fp);
			}
			
			cout << outStr << endl;
		}
	}

	return;
}

//------------------------------------------------------------------------------------------

void bSE_updateProgressMsg(char *msg)
{
	FILE			*fp = nil;
	
	if (gShowProgress)
	{
		if (msg)
		{
			fp = fopen(gProgressMsgPath, "w");
			if (fp) 
			{
				fwrite(msg, 1, strlen(msg), fp);
				fclose(fp);
			}
			
			cout << msg << endl;
		}
	}

	return;
}

//------------------------------------------------------------------------------------------

void bSE_updateDataTransferProgress(unsigned int transferredBytes)
{
	int	perc;
	
	if ((gShowProgress)&&(gTotalDataSize))
	{
		gProcessedDataSize += transferredBytes;
		
		perc = (int) ((gProcessedDataSize * 100L) / gTotalDataSize);
		if (perc > 100) perc = 100;
		bSE_updateProgress(perc, 100);
	}

	return;
}

//------------------------------------------------------------------------------------------

void bSE_updateDataDeleteProgress(void)
{
	int	perc;
	
	if ((gShowProgress)&&(gTotalCount))
	{
		++gProcessedCount;
		
		perc = (int) ((gProcessedCount * 100L) / gTotalCount);
		if (perc > 100) perc = 100;
		bSE_updateProgress(perc, 100);
	}

	return;
}

//------------------------------------------------------------------------------------------

int bSE_processShellCommands(struct shell_state *sh, char *cmdFile)
{
	int				error = 1;
	char			*text = nil, *pC, *cmdpC;
	int				totalCmdCount, doneCmdCount, textSize, ret;
	char			outStr[512];
	FILE			*fp = nil;
	
	fp = fopen(cmdFile, "r");
	if (! fp)
		goto finish;
		
	fseek( fp, 0L, SEEK_END );
	textSize = ftell( fp );
	text = (char *) calloc(1L, textSize+4L);
	if (! text)
		goto finish;
		
	fseek( fp, 0L, SEEK_SET );
	fread(text, 1, textSize, fp);
	fclose(fp);
	fp = nil;
	
	sprintf(outStr, "processing: %s", cmdFile);
	my_cout(outStr, kForceOut);
	
	if (gShowProgress) // count the commands
	{
		totalCmdCount = doneCmdCount = 0;

		pC = text;
		while(*pC)
			if (*pC++ == '\n')
				++totalCmdCount;
		
		sprintf(outStr, "commands: %d", totalCmdCount);
		my_cout(outStr, kForceOut);
		
		bSE_updateProgress(doneCmdCount, totalCmdCount);
	}
	
	cmdpC = text;
	while(1)
	{
		ret = SHELL_TERMINATE;
		
		bSE_updateProgress(++doneCmdCount, totalCmdCount);

		pC = strstr(cmdpC, "\n");
		if (pC)
		{
			*pC = 0;
			
			cout << cmdpC << endl;
			ret = exec_line(cmdpC, sh);
			
			++pC;
			cmdpC = pC;
		}
		
		if (bSE_abort()) // abort command processing
			break;
		
		if (ret != SHELL_CONTINUE)
			break;
	}
	
	error = 0;
	
finish:		
	
	if (text) free(text);
	text = nil;
	
	if (fp) fclose(fp);
	fp = nil;
	
	bSE_updateManifest(sh);
	return error;
}

//------------------------------------------------------------------------------------------

void bSE_deviceCommand(char *command, char *forSerialNo)
{
	struct shell_state				*sh;
	unsigned int					ret;
	string1K						outStr, serialNo;
	int								numDev = 0;
	UUTextParser					inputTP, progTP;
	char							*pC;

	// perform the command
	// if 'forSerialNo' is given, then perform the command on the device with that serial number
	// (which means the device must be in normal mode)
	
	ret = CFRunLoopRunInMode(kCFRunLoopDefaultMode, 0.2, false);
	switch(ret) // add marker strings to output
	{
		case kCFRunLoopRunFinished:			my_cout("kCFRunLoopRunFinished", kForceOut);		break;
		case kCFRunLoopRunStopped:			my_cout("kCFRunLoopRunStopped", kForceOut);			break;
		case kCFRunLoopRunTimedOut:			my_cout("//STATUS_TIMED_OUT//", kForceOut);			break;
		case kCFRunLoopRunHandledSource:	my_cout("kCFRunLoopRunHandledSource", kForceOut);	break;
	}
	
	//---
	
	for (int i = 0; i < kMaxDevices; ++i)
		if (gDEV[i].sh)
			++numDev;

	sprintf(outStr, "//NUM_TOTAL: %d//", numDev);
	my_cout(outStr, kForceOut);
			
	//---
	
	sprintf(outStr, "//COMMAND: %s//", command);
	my_cout(outStr, kForceOut);

	if (! strcmp(command, "LIST_DIRECTORY"))
	{
		struct shell_state	*theSH = nil;
		string256			theSN, theService;
		bool				theAfc2;
		bool				listDeep;
		
		listDeep = gConfigTP.mGetNumber("DEEP_LISTING", 0.0) ? true : false;

		if (gSkipListPath[0])
		{
			gSkipTP.mSetText(gSkipListPath);
			sprintf(outStr, "//SKIP_LIST %s//", gSkipListPath);
			my_cout(outStr, kForceOut);
		}
	
		strcpy(theService,	gConfigTP.mGetString("SRC_SERVICE", "AFC")); // or AFC2 or DISK
		strcpy(theSN, gConfigTP.mGetString("SRC_SN", ""));	
			
		theAfc2 = false;
		
		if (strcmp(theService, "DISK")) // find the source device, start the desired service, open a connection
		{
			if (! strcmp(theService, "AFC2"))
				theAfc2 = true;
				
			for (int i = 0; i < kMaxDevices; ++i)
				if (! bSE_openConnection(&gDEV[i], theSN, theAfc2))
				{
					theSH = gDEV[i].sh;
					break;
				}

			if (! theSH)
			{
				sprintf(outStr, "//ERROR//Target invalid - SN:%s AFC2:%d//", theSN, (theAfc2)? 1 : 0);
				my_cout(outStr, kForceOut);
				goto finish;
			}
		}
		
		sprintf(outStr, "//%s %s %s//", theService, theSN, gPath1);
		my_cout(outStr, kForceOut);
		
		if (gUseManifest)
			if (theSH)
				bSE_loadManifest(theSH);
		
		if (listDeep)
		{
			if (bSE_listDirectoryDeep(theSH, gPath1))
				ret = 1;
		}
		else
		{
			if (bSE_listDirectory(theSH, gPath1))
				ret = 1;
		}
		
		if (! ret)
		{	
			sprintf(outStr, "//OK//");
			my_cout(outStr, kForceOut);
		}
		
		gSkipTP.mDisposeText(); // taken care of!
		goto finish;
	}
	else if (! strcmp(command, "COUNT_DEVICES"))
	{
		int numMode[SHELL_LAST];
		
		MEMSET(numMode, 0, sizeof(numMode));
		
		for (int i = 0; i < kMaxDevices; ++i)
		{
			sh = gDEV[i].sh;
			if (! sh)
				continue;
			
			if ((sh->shell_mode >= 0)
				&&(sh->shell_mode < SHELL_LAST)
			)
			++numMode[ sh->shell_mode ];
			
			//---
			delete sh;
			MEMSET(&gDEV[i], 0, sizeof(bSEDevice));
		}
		
		sprintf(outStr, 
		"//NUM_NORMAL: %d//\n"
		"//NUM_RESTORE: %d//\n"
		"//NUM_RECOVERY: %d//\n"
		"//NUM_DFU: %d//\n"
		, numMode[SHELL_NORMAL]
		, numMode[SHELL_RESTORE]
		, numMode[SHELL_RECOVERY]
		, numMode[SHELL_DFU]
		);
		
		my_cout(outStr, kForceOut);
		my_cout("//OK//", kForceOut);
		goto finish;
	}
	else if (! strcmp(command, "DATA_INFO"))
	{
		struct shell_state	*theSH = nil;
		string256			theSN, theService;
		string1K			thePath;
		bool				theAfc2;
		char				*pathpC;
		
		gShowProgress	= gConfigTP.mGetNumber("SHOW_PROGRESS", 0.0) ? true : false;
		gCheckAbort		= gConfigTP.mGetNumber("CHECK_ABORT", 1.0) ? true : false;
		
		strcpy(theService,	gConfigTP.mGetString("SRC_SERVICE", "AFC")); // or AFC2 or DISK
		strcpy(theSN, gConfigTP.mGetString("SRC_SN", ""));	
			
		theAfc2 = false;
		
		if (strcmp(theService, "DISK")) // find the source device, start the desired service, open a connection
		{
			if (! strcmp(theService, "AFC2"))
				theAfc2 = true;
				
			for (int i = 0; i < kMaxDevices; ++i)
				if (! bSE_openConnection(&gDEV[i], theSN, theAfc2))
				{
					theSH = gDEV[i].sh;
					break;
				}

			if (! theSH)
			{
				sprintf(outStr, "//ERROR//Target invalid - SN:%s AFC2:%d//", theSN, (theAfc2)? 1 : 0);
				my_cout(outStr, kForceOut);
				goto finish;
			}
		}
		
		sprintf(outStr, "//%s %s//", theService, theSN);
		my_cout(outStr, kForceOut);
		
		// each input line *pair* is an absolute file or folder path on the device or on disk
		// followed by 'DISK' or 'DEVICE'
		
		inputTP.mSetText(gInputPath);
		if (! inputTP.fText)
		{
			sprintf(outStr, "//ERROR//Input invalid//");
			my_cout(outStr, kForceOut);
			goto finish;
		}
		
		// determine the total number of items to get info for
		// actually, this *is* getting the info :-)
		{
			UUTextParser			tmpTP;
			std::vector<bSEItem>	IL;
			bSEHdr					hdr;
			struct shell_state		*useSH;
			
			bSE_updateProgress(0, -1);
			
			MEMSET(&hdr, 0, sizeof(bSEHdr));
			gTotalCount = gProcessedCount = 0L;
			
			tmpTP.mSetText(inputTP.fText, inputTP.fTextSize);
			pathpC = tmpTP.fText;
			while(1)
			{
				if (bSE_abort()) // abort processing
				{
					ret = 99;
					break;
				}

				thePath[0] = 0;
				useSH = theSH;
				
				pC = strstr(pathpC, "\n");
				if (pC) // separator between the two paths
				{
					*pC = 0; // chop
					strcpy(thePath, pathpC); // get the source path
					
					++pC; // advance to start of dest path
					pathpC = pC; // remember the start
					
					pC = strstr(pathpC, "\n");
					if (pC) // end of line
					{
						*pC = 0; // chop
						
						if (strstr(pathpC, "DISK"))
							useSH = nil;

						++pC;
						pathpC = pC;
					}
				}
				else // done
					break;
					
				bSE_dataInfo(useSH, thePath, &IL);
			}
			
			bSE_writeBinaryOutput(&hdr, &IL);
		}
		
		if (! ret)
		{	
			sprintf(outStr, "//OK//");
			my_cout(outStr, kForceOut);
		}
		
		goto finish;
	}
	else if (! strcmp(command, "DATA_DELETE"))
	{
		struct shell_state	*theSH = nil;
		string256			theSN, theService;
		string1K			thePath;
		bool				theAfc2;
		char				*pathpC;
		
		gShowProgress	= gConfigTP.mGetNumber("SHOW_PROGRESS", 0.0) ? true : false;
		gCheckAbort		= gConfigTP.mGetNumber("CHECK_ABORT", 1.0) ? true : false;
		
		strcpy(theService,	gConfigTP.mGetString("SRC_SERVICE", "AFC")); // or AFC2 or DISK
		strcpy(theSN, gConfigTP.mGetString("SRC_SN", ""));	
			
		theAfc2 = false;
		
		if (strcmp(theService, "DISK")) // find the source device, start the desired service, open a connection
		{
			if (! strcmp(theService, "AFC2"))
				theAfc2 = true;
				
			for (int i = 0; i < kMaxDevices; ++i)
				if (! bSE_openConnection(&gDEV[i], theSN, theAfc2))
				{
					theSH = gDEV[i].sh;
					break;
				}

			if (! theSH)
			{
				sprintf(outStr, "//ERROR//Target invalid - SN:%s AFC2:%d//", theSN, (theAfc2)? 1 : 0);
				my_cout(outStr, kForceOut);
				goto finish;
			}
		}
		
		sprintf(outStr, "//%s %s//", theService, theSN);
		my_cout(outStr, kForceOut);
		
		// each input line is an absolute file or folder path on the device
		
		inputTP.mSetText(gInputPath);
		if (! inputTP.fText)
		{
			sprintf(outStr, "//ERROR//Input invalid//");
			my_cout(outStr, kForceOut);
			goto finish;
		}
		
		if (gShowProgress) // determine the total number of items to delete
		{
			UUTextParser	tmpTP;
			
			bSE_updateProgress(0, -1);
			
			gTotalCount = gProcessedCount = 0L;
			
			tmpTP.mSetText(inputTP.fText, inputTP.fTextSize);
			pathpC = tmpTP.fText;
			while(1)
			{
				if (bSE_abort()) // abort processing
				{
					ret = 99;
					break;
				}

				thePath[0] = 0;
				
				pC = strstr(pathpC, "\n");
				if (pC) // end of line
				{
					*pC = 0; // chop
					strcpy(thePath, pathpC); // get the path

					++pC;
					pathpC = pC;
				}
				else // done
					break;
					
				gTotalCount += bSE_totalCount(theSH, thePath);
			}
		}
		
		ret = 0;
		pathpC = inputTP.fText;
		while(1)
		{
			if (bSE_abort()) // abort processing
			{
				ret = 99;
				break;
			}

			thePath[0] = 0;
			
			pC = strstr(pathpC, "\n");
			if (pC) // end of line
			{
				*pC = 0; // chop
				strcpy(thePath, pathpC); // get the path

				++pC;
				pathpC = pC;
			}
			else // done
				break;
			
			if (bSE_dataDelete(theSH, thePath))
			{
				ret = 1;
				break; //?
			}
		}	
		
		if (gAppChange)
		{
			sprintf(gMsgStr, "//APP_CHANGE//");
			my_cout(gMsgStr, kForceOut);
		}
		
		if (! ret)
		{	
			sprintf(outStr, "//OK//");
			my_cout(outStr, kForceOut);
		}
		
		goto finish;
	}
	else if (! strcmp(command, "DATA_RENAME"))
	{
		struct shell_state	*theSH = nil;
		string256			theSN, theService;
		string1K			thePath, theName;
		bool				theAfc2;
		
		strcpy(theService,	gConfigTP.mGetString("SRC_SERVICE", "AFC")); // or AFC2 or DISK
		strcpy(theSN, gConfigTP.mGetString("SRC_SN", ""));	
		
		strcpy(thePath, gConfigTP.mGetString("PATH1", ""));	
		strcpy(theName, gConfigTP.mGetString("NAME1", ""));	
			
		theAfc2 = false;
		
		if (strcmp(theService, "DISK")) // find the source device, start the desired service, open a connection
		{
			if (! strcmp(theService, "AFC2"))
				theAfc2 = true;
				
			for (int i = 0; i < kMaxDevices; ++i)
				if (! bSE_openConnection(&gDEV[i], theSN, theAfc2))
				{
					theSH = gDEV[i].sh;
					break;
				}

			if (! theSH)
			{
				sprintf(outStr, "//ERROR//Target invalid - SN:%s AFC2:%d//", theSN, (theAfc2)? 1 : 0);
				my_cout(outStr, kForceOut);
				goto finish;
			}
		}
		
		sprintf(outStr, "//%s %s %s -> %s//", theService, theSN, thePath, theName);
		my_cout(outStr, kForceOut);
		
		ret = bSE_dataRename(theSH, thePath, theName);
		if (! ret)
		{	
			sprintf(outStr, "//OK//");
			my_cout(outStr, kForceOut);
		}
		
		goto finish;
	}
	else if (! strcmp(command, "DATA_TRANSFER"))
	{
		struct shell_state	*srcSH = nil;
		struct shell_state	*dstSH = nil;
		string256			srcSN, dstSN, srcService, dstService;
		string1K			srcPath, dstPath;
		bool				srcAfc2, dstAfc2, sameDev = false;
		int					transferFlags = kDT_default;
		char				*pathpC;
		
		gShowProgress	= gConfigTP.mGetNumber("SHOW_PROGRESS", 0.0) ? true : false;
		gCheckAbort		= gConfigTP.mGetNumber("CHECK_ABORT", 1.0) ? true : false;
		
		if (gConfigTP.mGetNumber("FLAT", 1.0))
			transferFlags |= kDT_copyFlat;
		
		if (gConfigTP.mGetNumber("MERGE", 0.0))
			transferFlags |= kDT_copyMerge;
		
		if (gConfigTP.mGetNumber("CONVERT_PLISTS", 0.0))
			transferFlags |= kDT_convertPlists;
		
		if (gConfigTP.mGetNumber("CONVERT_PNG", 0.0))
			transferFlags |= kDT_convertPng;
		
		strcpy(srcService,	gConfigTP.mGetString("SRC_SERVICE", "AFC")); // or AFC2 or DISK
		strcpy(dstService,	gConfigTP.mGetString("DST_SERVICE", "AFC")); // or AFC2 or DISK
		
		strcpy(srcSN, gConfigTP.mGetString("SRC_SN", ""));			
		strcpy(dstSN, gConfigTP.mGetString("DST_SN", ""));			
			
		srcAfc2 = dstAfc2 = false;
		
		if (strcmp(srcService, "DISK")) // find the source device, start the desired service, open a connection
		{
			if (! strcmp(srcService, "AFC2"))
				srcAfc2 = true;
				
			for (int i = 0; i < kMaxDevices; ++i)
				if (! bSE_openConnection(&gDEV[i], srcSN, srcAfc2))
				{
					srcSH = gDEV[i].sh;
					break;
				}

			if (! srcSH)
			{
				sprintf(outStr, "//ERROR//Source invalid - SN:%s AFC2:%d//", srcSN, (srcAfc2)? 1 : 0);
				my_cout(outStr, kForceOut);
				goto finish;
			}
		}
		
		if (strcmp(dstService, "DISK")) // find the destination device, start the desired service, open a connection
		{
			if (! strcmp(dstService, "AFC2"))
				dstAfc2 = true;
			
			if (! strcmp(srcSN, dstSN))
			{
				dstSH	= srcSH;
				dstAfc2	= srcAfc2;
				sameDev	= true;
			}
			else
			{
				for (int i = 0; i < kMaxDevices; ++i)
					if (! bSE_openConnection(&gDEV[i], dstSN, dstAfc2))
					{
						dstSH = gDEV[i].sh;
						break;
					}
			}
			
			if (! dstSH)
			{
				sprintf(outStr, "//ERROR//Destination invalid - SN:%s AFC2:%d//", dstSN, (dstAfc2)? 1 : 0);
				my_cout(outStr, kForceOut);
				goto finish;
			}
		}
		
		if (gUseManifest)
		{
			if (srcSH)
				bSE_loadManifest(srcSH);
			
			if ((dstSH)&&(! sameDev))
				bSE_loadManifest(dstSH);
		}
		
		sprintf(outStr, "//%s %s ==> %s %s//", srcService, srcSN, dstService, dstSN);
		my_cout(outStr, kForceOut);
		
		// input is a source and a destination line *pair*
		// of absolute file or folder paths on the respective device
		
		inputTP.mSetText(gInputPath);
		if (! inputTP.fText)
		{
			sprintf(outStr, "//ERROR//Input invalid//");
			my_cout(outStr, kForceOut);
			goto finish;
		}
		
		if (gSkipListPath[0])
		{
			gSkipTP.mSetText(gSkipListPath);
			sprintf(outStr, "//SKIP_LIST %s//", gSkipListPath);
			my_cout(outStr, kForceOut);
		}
	
		gTotalDataSize = gProcessedDataSize = 0L;
		bSE_updateProgress(0, -1);
		bSE_updateProgressMsg("[1]");
				
		#if 1 
		{
			// do a preflight to make a list of items to transfer
			// produce a list of src -> dest items
			// also calculate the total data size for progress bar
			
			gDataTransferPreflight = true;
		
			ret = 0;
			pathpC = inputTP.fText;
			while(1)
			{
				if (bSE_abort()) // abort processing
				{
					ret = 99;
					break;
				}

				srcPath[0] = dstPath[0] = 0;
				
				pC = strstr(pathpC, "\n");
				if (pC) // separator between the two paths
				{
					*pC = 0; // chop
					strcpy(srcPath, pathpC); // get the source path
					
					++pC; // advance to start of dest path
					pathpC = pC; // remember the start
					
					pC = strstr(pathpC, "\n");
					if (pC) // end of line
					{
						*pC = 0; // chop
						strcpy(dstPath, pathpC); // get the source path

						++pC;
						pathpC = pC;
					}
				}
				else // done
					break;
				
				if (bSE_dataTransfer(srcSH, srcPath, dstSH, dstPath, transferFlags))
				{
					ret = 1;
					break; //?
				}
			}
			
			gDataTransferPreflight = false;
			gSkipTP.mDisposeText(); // taken care of!
			bSE_updateProgressMsg("[2]");
			
			if (! ret) // now do the actual transfer
			{
				bSEItemDT	*pDT;
				
				// determine the total data size for the progress bar
				
				pDT = &gDataTransferIL[0];
				for (int i = 0; i < (int) gDataTransferIL.size(); ++i, ++pDT)
					gTotalDataSize += pDT->srcSize;
				
				transferFlags |= kDT_copyFlat; //!
				
				pDT = &gDataTransferIL[0];
				for (int i = 0; i < (int) gDataTransferIL.size(); ++i, ++pDT)
				{
					if ((i % 5) == 0)		// limit the check frequency
						if (bSE_abort())	// abort processing
						{
							ret = 99;
							break;
						}
					
					bSE_updateProgressMsg(pDT->srcPath);
					
					if (pDT->srcIsDir)	ret = bSE_copyFolder(pDT->srcSH, pDT->srcPath, pDT->dstSH, pDT->dstPath, transferFlags);
					else				ret = bSE_copyFile(pDT->srcSH, pDT->srcPath, pDT->dstSH, pDT->dstPath, transferFlags);
					
					/* what if files disappear *while* we're going thru this list? - keep going
					if (ret)
						break;
					*/
				}
			}
		}
		#else
		{
			if (gShowProgress) // determine the total data size to transfer
			{
				UUTextParser	tmpTP;
				
				tmpTP.mSetText(inputTP.fText, inputTP.fTextSize);
				pathpC = tmpTP.fText;
				while(1)
				{
					if (bSE_abort()) // abort processing
					{
						ret = 99;
						break;
					}

					srcPath[0] = 0;
					
					pC = strstr(pathpC, "\n");
					if (pC) // separator between the two paths
					{
						*pC = 0; // chop
						strcpy(srcPath, pathpC); // get the source path
						
						++pC; // advance to start of dest path
						pC = strstr(pC, "\n");
						if (pC) // end of line
						{
							++pC;
							pathpC = pC;
						}
					}
					else // done
						break;
						
					gTotalDataSize += bSE_totalDataSize(srcSH, srcPath, transferFlags);
				}
			}
			
			//---
			
			bSE_updateProgressMsg("[2]");
			
			ret = 0;
			pathpC = inputTP.fText;
			while(1)
			{
				if (bSE_abort()) // abort processing
				{
					ret = 99;
					break;
				}

				srcPath[0] = dstPath[0] = 0;
				
				pC = strstr(pathpC, "\n");
				if (pC) // separator between the two paths
				{
					*pC = 0; // chop
					strcpy(srcPath, pathpC); // get the source path
					
					++pC; // advance to start of dest path
					pathpC = pC; // remember the start
					
					pC = strstr(pathpC, "\n");
					if (pC) // end of line
					{
						*pC = 0; // chop
						strcpy(dstPath, pathpC); // get the source path

						++pC;
						pathpC = pC;
					}
				}
				else // done
					break;
				
				if (bSE_dataTransfer(srcSH, srcPath, dstSH, dstPath, transferFlags))
				{
					ret = 1;
					break; //?
				}
			}	
		}
		#endif
		
		if (gAppChange)
		{
			sprintf(gMsgStr, "//APP_CHANGE//");
			my_cout(gMsgStr, kForceOut);
		}
		
		if (! ret)
		{	
			sprintf(outStr, "//OK//");
			my_cout(outStr, kForceOut);
		}
		
		goto finish;
	}
	else if (! strcmp(command, "ENTER_RECOVERY"))
	{
		bool found;
		
		for (int i = 0; i < kMaxDevices; ++i)
		{
			sh = gDEV[i].sh;
			if (! sh)
				continue;
			
			if (sh->shell_mode == SHELL_RECOVERY)
			{
				gWaitForRecoveryIndex = i;
				break;
			}
				
			if (sh->shell_mode != SHELL_NORMAL)
				continue;
			
			if (forSerialNo)
				if (forSerialNo[0])
				{
					found = false;	

					AMDeviceIsPaired(sh->dev);
					if (! AMDeviceValidatePairing(sh->dev))
						if (! AMDeviceStartSession(sh->dev))
						{
							bSE_getSerialNo(sh->dev, serialNo);
							if (! strcmp(forSerialNo, serialNo))
								found = true;
						}
					
					if (! found)
						continue;
				}
				
			my_cout("AMDeviceEnterRecovery: ", AMDeviceEnterRecovery(sh->dev));
			my_cout("//OK//", kForceOut);
			
			gWaitForRecoveryIndex = i;
			break;
		}
		
		goto finish;
	}
	else if (! strcmp(command, "COPY_DFU_FILE"))
	{
		CFStringRef file	= CFStringCreateWithCString(NULL, "WTF.s5l8900xall.RELEASE.dfu", kCFStringEncodingMacRoman);
		CFStringRef cmd_go	= CFStringCreateWithCString(NULL, "go", kCFStringEncodingMacRoman);
		
		for (int i = 0; i < kMaxDevices; ++i)
		{
			sh = gDEV[i].sh;
			if (! sh)
				continue;
		
			if (sh->shell_mode != SHELL_RECOVERY)
				continue;
			
			if (gWaitForRecoveryIndex != -1)
				if (i != gWaitForRecoveryIndex)
					continue;
					
			ret = sendFileToDevice(sh->recovery_dev, file);
			my_cout("sendFileToDevice: ", ret);
			
			if (! ret)
			{
				ret = sendCommandToDevice(sh->recovery_dev, cmd_go);
				my_cout("sendCommandToDevice: ", ret);
			}
			
			if (! ret) 
				my_cout("//OK//", kForceOut);
		}
		
		CFRelease(file);
		CFRelease(cmd_go);
		goto finish;
	}
	else if (! strcmp(command, "REBOOT")) // POINTLESS - DOES NOT EXIT RECOVERY MODE!
	{
		CFStringRef cmd_saveenv		= CFStringCreateWithCString(NULL, "saveenv", kCFStringEncodingMacRoman);
		CFStringRef cmd_reboot		= CFStringCreateWithCString(NULL, "reboot", kCFStringEncodingMacRoman);

		for (int i = 0; i < kMaxDevices; ++i)
		{
			sh = gDEV[i].sh;
			if (! sh)
				continue;
		
			if ((sh->shell_mode != SHELL_RECOVERY)
		//		&&(sh->shell_mode != SHELL_DFU)	crashes
			)
			continue;
				
			ret = sendCommandToDevice(sh->recovery_dev, cmd_saveenv);
			my_cout("sendCommandToDevice: ", ret);
			
			if (! ret)
			{
				ret = sendCommandToDevice(sh->recovery_dev, cmd_reboot);
				my_cout("sendCommandToDevice: ", ret);
			}
			
			if (! ret) 
				my_cout("//OK//", kForceOut);
		}
		
		CFRelease(cmd_saveenv);
		CFRelease(cmd_reboot);
		goto finish;
	}
	else if (! strcmp(command, "EXIT_RECOVERY")) // NOT RECOMMENDED!
	{
		CFStringRef cmd_autoboot	= CFStringCreateWithCString(NULL, "setenv auto-boot true", kCFStringEncodingMacRoman);
		CFStringRef cmd_saveenv		= CFStringCreateWithCString(NULL, "saveenv", kCFStringEncodingMacRoman);
		CFStringRef cmd_fsboot		= CFStringCreateWithCString(NULL, "fsboot", kCFStringEncodingMacRoman);

		// causes segmentation fault for DFU mode
		// may cause problems for RECOVERY mode (seems to work, but there were reports of recovery mode after device is shut off)
		// safer to use another method!
		
		for (int i = 0; i < kMaxDevices; ++i)
		{
			sh = gDEV[i].sh;
			if (! sh)
				continue;
		
			if (sh->shell_mode != SHELL_RECOVERY)
				continue;
			
			ret = sendCommandToDevice(sh->recovery_dev, cmd_autoboot);
			my_cout("sendCommandToDevice: ", ret);
			
			if (! ret)
			{
				ret = sendCommandToDevice(sh->recovery_dev, cmd_saveenv);
				my_cout("sendCommandToDevice: ", ret);
			}
			
			if (! ret)
			{
				ret = sendCommandToDevice(sh->recovery_dev, cmd_fsboot);
				my_cout("sendCommandToDevice: ", ret);
			}
			
			if (! ret) 
				my_cout("//OK//", kForceOut);
		}
		
		CFRelease(cmd_autoboot);
		CFRelease(cmd_saveenv);
		CFRelease(cmd_fsboot);
		goto finish;
	}
	else if (! strcmp(command, "DEVICE_INFO"))
	{
		for (int i = 0; i < kMaxDevices; ++i)
		{
			sh = gDEV[i].sh;
			if (! sh)
				continue;
			
			if (sh->shell_mode == SHELL_RESTORE)
			{
				sprintf(outStr, "//DEVICEMODE%d: RESTORE//", sh->devNo);
				my_cout(outStr, kForceOut);
				//---
			}
			else if (sh->shell_mode == SHELL_RECOVERY)
			{
				sprintf(outStr, "//DEVICEMODE%d: RECOVERY//", sh->devNo);
				my_cout(outStr, kForceOut);
				//---
			}
			else if (sh->shell_mode == SHELL_DFU)
			{
				sprintf(outStr, "//DEVICEMODE%d: DFU//", sh->devNo);
				my_cout(outStr, kForceOut);
				//---
			}
			else if (sh->shell_mode == SHELL_NORMAL)
			{
				sprintf(outStr, "//DEVICEMODE%d: NORMAL//", sh->devNo);
				my_cout(outStr, kForceOut);
				//---
				
				ret = bSE_openConnection(&gDEV[i], forSerialNo, false);
				if (ret)
					continue;
				
				if (forSerialNo)
					if (forSerialNo[0])
						sh->devNo = 0; // so we can parse it for the single device!
					
				bSE_getDeviceInfo(sh);

				AFCConnectionClose(sh->conn);
			}
		}
		
		my_cout("//OK//", kForceOut);
		goto finish;
	}
	else
	{
		my_cout("//ERROR: Unknown Command//", kForceOut);
		goto finish;
	}
	
finish:	
	return;
}

//------------------------------------------------------------------------------------------
#if 0 
#pragma mark -
#endif
//------------------------------------------------------------------------------------------

errcode mConvertPlist(char *data, int dataSize, UUTextParser *outTP)
{
	errcode			err = 0;
	FILE			*fp;
	string512		tmpPath, tmpPath2, outputPath;
	string1K		shellCmd, shellCmd2, result, redirect;
	
	outTP->mDisposeText();
	
	sprintf(tmpPath,	"/tmp/tmpIn%d.plist", gInstanceID);
	sprintf(tmpPath2,	"/tmp/tmpOut%d.plist", gInstanceID);
	sprintf(outputPath,	"/tmp/plutilOut%d.txt", gInstanceID);
	
	fp = fopen(tmpPath, "wb" );
	if (! fp) {err = 1; goto finish;}
	fwrite(data, 1, dataSize, fp);
	fclose(fp);
	
	//---
	
	sprintf(shellCmd, "plutil -convert xml1 '%s' -o '%s'", tmpPath, tmpPath2);
	
	sprintf(redirect, ">& '%s'", outputPath); // redirects stdout and stderr(!) to the given text file
	sprintf(shellCmd2, "tcsh -c \"%s %s\"", shellCmd, redirect); // wrap the shell invocation, quotes and the redirect around the command!
	mShellCommand(shellCmd2, result, sizeof(result)-1);
	
	//---
	
	err = outTP->mSetText(tmpPath2);
	
finish:
	unlink(tmpPath);
	unlink(tmpPath2);
	unlink(outputPath);
	return err;
}

//------------------------------------------------------------------------------------------

errcode mConvertPng(char *pngPath)
{
	errcode			err = 0;
	string512		outputPath;
	string1K		shellCmd, shellCmd2, result, redirect;
	
	sprintf(outputPath,	"/tmp/fixpngOut%d.txt", gInstanceID);
	
	sprintf(shellCmd, "./fixpng -f '%s' '%s'", pngPath, pngPath);
	
	sprintf(redirect, ">& '%s'", outputPath); // redirects stdout and stderr(!) to the given text file
	sprintf(shellCmd2, "tcsh -c \"%s %s\"", shellCmd, redirect); // wrap the shell invocation, quotes and the redirect around the command!
	mShellCommand(shellCmd2, result, sizeof(result)-1);
	
finish:
	unlink(outputPath);
	return err;
}

//------------------------------------------------------------------------------------------

errcode mExitRecoveryWithIPhonecomm(char *commPath)
{
	errcode			err = 0;
	string512		outputPath;
	string1K		shellCmd, shellCmd2, result, redirect;
	
	strcpy(outputPath,	"/tmp/iphonecommOut.txt");
	
	sprintf(shellCmd, "cd '%s'; ./iphonecomm -s 4", commPath);
	
	sprintf(redirect, ">& '%s'", outputPath); // redirects stdout and stderr(!) to the given text file
	sprintf(shellCmd2, "tcsh -c \"%s %s\"", shellCmd, redirect); // wrap the shell invocation, quotes and the redirect around the command!
	mShellCommand(shellCmd2, result, sizeof(result)-1);
	
finish:
	unlink(outputPath);
	return err;
}

//------------------------------------------------------------------------------------------

errcode mShellCommand(char *command, char *result, long maxResultSize)
{
	errcode			err = 0;
	FILE			*fp = nil;
	int				bytesRead, t = 0, retCode = 0;

	result[0] = 0;

	fp = popen( command, "r" );
	if (! fp)
	{err = 1; goto finish;}
	
	t = setlinebuf(fp);
	
	bytesRead = fread( result, sizeof(char), maxResultSize, fp); // stdout gets redirected to this FILE*, stderr still goes to the console
	result[bytesRead] = 0;
	
	retCode = pclose(fp);
	fp = nil;
		
finish:
	return retCode;
}

//----------------------------------------------------------------------------------------------

int mGetUnicodeString(char *cs, UniChar *ucs)
{
	errcode 			err = 0;
	int					length = 0;
	TextToUnicodeInfo	tucinfo = nil;
	Str255				pstr;
	ByteCount			uclen;
	
	// returns the given c string in unicode format
	// assumes that 'ucs' holds at least 256 characters
	
	ucs[0] = 0;
	c2pstrcpy(pstr, cs);
	
	err = CreateTextToUnicodeInfoByEncoding(kUnicodeUTF8Format, &tucinfo);
	if (err) goto finish;

	err = ConvertFromPStringToUnicode(tucinfo, pstr, 256L, &uclen, ucs);
	if (err) goto finish;
	
	length = uclen/sizeof(UniChar);
	
finish:
	if (tucinfo) DisposeTextToUnicodeInfo(&tucinfo);
	return length;
}	

//------------------------------------------------------------------------------------------

errcode mSplitPath(const char *fullPath, char *path, char *name)
{
	errcode		err = 0;
	string512	thePath, theName;
	short		n;
	char		*pC;
	
	// split 'fullPath' into 'path' and 'name'
	
	if (fullPath)
	{
		strcpy(theName, fullPath);
		strcpy(thePath, fullPath);
	}
	
	if (path) path[0] = 0;
	if (name) name[0] = 0;
	
	if (! fullPath)
		goto finish;
		
	// find the name
	
	n = strlen(thePath);
	if (! n) goto finish;

	if (thePath[n-1] == '/')
	{
		thePath[n-1] = 0;
		--n;
		
		if (! n) goto finish;
	}
	
	pC = (char *) &thePath[n-1];
	
	for (int i = n; i > 0; --i, --pC)
	{
		if (*pC == '/')
		{
			strcpy(theName, pC+1);
			pC[1] = 0; // this takes care of thePath
			break;
		}
	}
	
	if (path) strcpy(path, thePath);
	if (name) strcpy(name, theName);
	
finish:
	return err;
}

//------------------------------------------------------------------------------------------

errcode mGetSoundInfo_AAC(char *path, bSEItem *pI)
{
	errcode			err = kUUERR_invalid;
	string1K		shellCmd, shellCmd2, result, line;
	UUTextParser	TP;
	
	if ((! strstr(path, ".m4a"))
		&&(! strstr(path, ".m4p"))
	)
	goto finish;

	sprintf(shellCmd, "./AtomicParsley '%s' -t", path);
	sprintf(shellCmd2, "tcsh -c \"%s\"", shellCmd);
	mShellCommand(shellCmd2, result, sizeof(result)-1);

	//sprintf(gMsgStr, "//mGetSoundInfo_AAC//%s//\n%s", path, result);
	//my_cout(gMsgStr, kForceOut);
	
	// parse the output
	
	TP.mSetText(result, strlen(result));
	
	TP.mSetPos(0);
	if (! TP.mGetStringBetween("nam\" contains: ", "\n", line, sizeof(line)-1))
		strncpy(pI->soundTitle, line, sizeof(pI->soundTitle)-1);

	TP.mSetPos(0);
	if (! TP.mGetStringBetween("ART\" contains: ", "\n", line, sizeof(line)-1))
		strncpy(pI->soundArtist, line, sizeof(pI->soundArtist)-1);
	
	if ((pI->soundTitle[0])||(pI->soundArtist[0]))
		err = kUUERR_noErr;

finish:	
	return err;
}

//------------------------------------------------------------------------------------------

errcode mGetSoundInfo_MP3(char *path, bSEItem *pI)
{
	errcode		err = kUUERR_invalid;
	
	if (! strstr(path, ".mp3"))
		goto finish;

{	
	TagLib::FileRef tf(path);

	if (!tf.isNull() && tf.tag()) 
	{
		TagLib::Tag	*tag = tf.tag();
	  
		strncpy(pI->soundTitle, tag->title().toCString(true), sizeof(pI->soundTitle)-1);
		strncpy(pI->soundArtist, tag->artist().toCString(true), sizeof(pI->soundArtist)-1);
		
		if ((pI->soundTitle[0])||(pI->soundArtist[0]))
			err = kUUERR_noErr;
	}
	
	/*
	if (!tf.isNull() && tf.audioProperties()) 
	{
		TagLib::AudioProperties *properties = tf.audioProperties();

		pI->soundDuration = properties->length();
	}
	*/
}
	
	/*
	if (!f.isNull() && f.tag()) 
	{
	  TagLib::Tag *tag = f.tag();

	  cout << "-- TAG --" << endl;
	  cout << "title   - \"" << tag->title()   << "\"" << endl;
	  cout << "artist  - \"" << tag->artist()  << "\"" << endl;
	  cout << "album   - \"" << tag->album()   << "\"" << endl;
	  cout << "year    - \"" << tag->year()    << "\"" << endl;
	  cout << "comment - \"" << tag->comment() << "\"" << endl;
	  cout << "track   - \"" << tag->track()   << "\"" << endl;
	  cout << "genre   - \"" << tag->genre()   << "\"" << endl;
	}

	if (!f.isNull() && f.audioProperties()) 
	{
	  TagLib::AudioProperties *properties = f.audioProperties();

	  int seconds = properties->length() % 60;
	  int minutes = (properties->length() - seconds) / 60;

	  cout << "-- AUDIO --" << endl;
	  cout << "bitrate     - " << properties->bitrate() << endl;
	  cout << "sample rate - " << properties->sampleRate() << endl;
	  cout << "channels    - " << properties->channels() << endl;
	  cout << "length      - " << minutes << ":" << formatSeconds(seconds) << endl;
	}
	*/

finish:	
	return err;
}

//------------------------------------------------------------------------------------------

